# (C) Copyright IBM Corp. 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
import requests
import json
import re
import logging
from lomond import WebSocket
from ibm_watson_machine_learning.utils import print_text_header_h1, print_text_header_h2, TRAINING_RUN_DETAILS_TYPE, STR_TYPE, STR_TYPE_NAME, docstring_parameter, str_type_conv, meta_props_str_conv, group_metrics, StatusLogger
import time
from ibm_watson_machine_learning.metanames import TrainingConfigurationMetaNames, TrainingConfigurationMetaNamesCp4d30
from ibm_watson_machine_learning.wml_client_error import WMLClientError, ApiRequestFailure
from ibm_watson_machine_learning.href_definitions import is_uid
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_boto3.exceptions import Boto3Error

logging.getLogger('lomond').setLevel(logging.CRITICAL)


class Training(WMLResource):
    """
       Train new models.
    """

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)
        self._ICP = client.ICP
        self.ConfigurationMetaNames = TrainingConfigurationMetaNames()
        if self._client.ICP_30 or self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            self.ConfigurationMetaNames = TrainingConfigurationMetaNamesCp4d30()

    @staticmethod
    def _is_training_uid(s):
        res = re.match('p\-[a-zA-Z0-9\-\_]+', s)
        return res is not None

    @staticmethod
    def _is_training_url(s):
        res = re.match('\/v3\/models\/p\-[a-zA-Z0-9\-\_]+', s)
        return res is not None

    def _is_model_definition_url(self, s):
        res = re.match('\/v2\/assets\/p\-[a-zA-Z0-9\-\_]+', s)
        return res is not None

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_status(self, training_uid):
        """
                Get the status of a training created.

                **Parameters**

                .. important::
                    #. **training_uid**:  training UID\n
                       **type**: str\n

                **Output**

                .. important::
                    **returns**: training_status\n
                    **return type**: dict\n

                **Example**

                 >>> training_status = client.training.get_status(training_uid)
        """

        training_uid = str_type_conv(training_uid)
        Training._validate_type(training_uid, 'training_uid', STR_TYPE, True)

        details = self.get_details(training_uid)

        if details is not None:
            return WMLResource._get_required_element_from_dict(details, u'details', [u'entity', u'status'])
        else:
            raise WMLClientError(u'Getting trained model status failed. Unable to get model details for training_uid: \'{}\'.'.format(training_uid))

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_details(self, training_uid=None, limit=None, asynchronous=False, get_all=False):
        """
        Get metadata of training(s). If training_uid is not specified returns all model spaces metadata.

        **Parameters**

        .. important::
            #. **training_uid**: Unique Id of Training (optional)\n
               **type**: str\n
            #. **limit**:  limit number of fetched records (optional)\n
               **type**: int\n
            #. **asynchronous**:  if True, it will work as a generator (optional)\n
               **type**: bool\n
            #. **get_all**:  if True, it will get all entries in 'limited' chunks (optional)\n
               **type**: bool\n

        **Output**

        .. important::
            **returns**: metadata of training(s)\n
            **return type**: dict\n
            The output can be {"resources": [dict]} or a dict\n

        .. note::
            If training_uid is not specified, all trainings metadata is fetched\n

        **Example**

         >>> training_run_details = client.training.get_details(training_uid)
         >>> training_runs_details = client.training.get_details()
         >>> training_runs_details = client.training.get_details(limit=100)
         >>> training_runs_details = client.training.get_details(limit=100, get_all=True)
         >>> training_runs_details = []
         >>> for entry in client.training.get_details(limit=100, asynchronous=True, get_all=True):
         >>>    training_runs_details.extend(entry)
        """
        ##For CP4D, check if either spce or project ID is set
        self._client._check_if_either_is_set()
        training_uid = str_type_conv(training_uid)
        Training._validate_type(training_uid, 'training_uid', STR_TYPE, False)

        url = self._client.service_instance._href_definitions.get_trainings_href()

        if training_uid is None:
            return self._get_artifact_details(url, training_uid, limit, 'trained models', _async=asynchronous, _all=get_all)

        else:
            return self._get_artifact_details(url, training_uid, limit, 'trained models')

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_href(training_details):
        """
            Get training_href from training details.

            **Parameters**

            .. important::
                #. **training_details**:  Metadata of the training created\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: training href\n
                **return type**: str

            **Example**

             >>> training_details = client.training.get_details(training_uid)
             >>> run_url = client.training.get_href(training_details)
        """

        Training._validate_type(training_details, u'training_details', object, True)
        if 'id' in training_details.get('metadata'):
            training_id = WMLResource._get_required_element_from_dict(training_details, u'training_details',
                                                                      [u'metadata', u'id'])
            return "/ml/v4/trainings/"+training_id
        else:
            Training._validate_type_of_details(training_details, TRAINING_RUN_DETAILS_TYPE)
            return WMLResource._get_required_element_from_dict(training_details, u'training_details', [u'metadata', u'href'])

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_uid(training_details):
        """
            Get training_uid from training details.

            **Parameters**

            .. important::
                #. **training_details**:  Metadata of the training created\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: Unique id of training\n
                **return type**: str

            **Example**

             >>> training_details = client.training.get_details(training_uid)
             >>> training_uid = client.training.get_uid(training_details)

        """

        Training._validate_type(training_details, u'training_details', object, True)
        return WMLResource._get_required_element_from_dict(training_details, u'training_details', [u'metadata', u'guid'])

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_id(training_details):
        """
            Get training_id from training details.

            **Parameters**

            .. important::
                #. **training_details**:  Metadata of the training created\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: Unique id of training\n
                **return type**: str

            **Example**

             >>> training_details = client.training.get_details(training_id)
             >>> training_id = client.training.get_id(training_details)

        """

        Training._validate_type(training_details, u'training_details', object, True)
        return WMLResource._get_required_element_from_dict(training_details, u'training_details',
                                                           [u'metadata', u'id'])

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def run(self, meta_props, asynchronous=True):
        """
        Create a new Machine Learning training.

        **Parameters**

        .. important::

            #. **meta_props**:  meta data of the training configuration. To see available meta names use:\n
               >>> client.training.ConfigurationMetaNames.show()

               **type**: str\n

            #. **asynchronous**:\n
               * True  - training job is submitted and progress can be checked later.\n
               * False - method will wait till job completion and print training stats.\n
               **type**: bool\n


        **Output**

        .. important::

            **returns**: Metadata of the training created\n
            **return type**: dict\n

        **Examples**
         Example meta_props for Training run creation in IBM Cloud Pak® for Data version 3.0.1 or above:\n
         >>> metadata = {
         >>>  client.training.ConfigurationMetaNames.NAME: 'Hand-written Digit Recognition',
         >>>  client.training.ConfigurationMetaNames.DESCRIPTION: 'Hand-written Digit Recognition Training',
         >>>  client.training.ConfigurationMetaNames.PIPELINE: {
         >>>                "id": "4cedab6d-e8e4-4214-b81a-2ddb122db2ab",
         >>>                "rev": "12",
         >>>                "model_type": "string",
         >>>                "data_bindings": [
         >>>                  {
         >>>                    "data_reference_name": "string",
         >>>                    "node_id": "string"
         >>>                  }
         >>>                ],
         >>>                "nodes_parameters": [
         >>>                  {
         >>>                    "node_id": "string",
         >>>                    "parameters": {}
         >>>                  }
         >>>                ],
         >>>                "hardware_spec": {
         >>>                  "id": "4cedab6d-e8e4-4214-b81a-2ddb122db2ab",
         >>>                  "rev": "12",
         >>>                  "name": "string",
         >>>                  "num_nodes": "2"
         >>>                }
         >>>      },
         >>>  client.training.ConfigurationMetaNames.TRAINING_DATA_REFERENCES: [{
         >>>          'type': 's3',
         >>>          'connection': {},
         >>>          'location': {
         >>>              'href': 'v2/assets/asset1233456',
         >>>          }
         >>>          "schema": "{ \"id\": \"t1\", \"name\": \"Tasks\", \"fields\": [ { \"name\": \"duration\", \"type\": \"number\" } ]}"
         >>>      }],
         >>> client.training.ConfigurationMetaNames.TRAINING_RESULTS_REFERENCE: {
         >>>          'id' : 'string',
         >>>          'connection': {
         >>>              'endpoint_url': 'https://s3-api.us-geo.objectstorage.service.networklayer.com',
         >>>              'access_key_id': '***',
         >>>              'secret_access_key': '***'
         >>>          },
         >>>          'location': {
         >>>              'bucket': 'wml-dev-results',
         >>>               'path' : "path"
         >>>          }
         >>>          'type': 's3'
         >>>      }
         >>>   }

         NOTE:  You can provide either one of the below values can be provided for training:\n
         * client.training.ConfigurationMetaNames.EXPERIMENT\n
         * client.training.ConfigurationMetaNames.PIPELINE\n
         * client.training.ConfigurationMetaNames.MODEL_DEFINITION\n
         Example meta_prop values for  training run creation in other versions:

         >>> metadata = {
         >>>  client.training.ConfigurationMetaNames.NAME: 'Hand-written Digit Recognition',
         >>>  client.training.ConfigurationMetaNames.TRAINING_DATA_REFERENCES: [{
         >>>          'connection': {
         >>>              'endpoint_url': 'https://s3-api.us-geo.objectstorage.service.networklayer.com',
         >>>              'access_key_id': '***',
         >>>              'secret_access_key': '***'
         >>>          },
         >>>          'source': {
         >>>              'bucket': 'wml-dev',
         >>>          }
         >>>          'type': 's3'
         >>>      }],
         >>> client.training.ConfigurationMetaNames.TRAINING_RESULTS_REFERENCE: {
         >>>          'connection': {
         >>>              'endpoint_url': 'https://s3-api.us-geo.objectstorage.service.networklayer.com',
         >>>              'access_key_id': '***',
         >>>              'secret_access_key': '***'
         >>>          },
         >>>          'target': {
         >>>              'bucket': 'wml-dev-results',
         >>>          }
         >>>          'type': 's3'
         >>>      },
         >>> client.training.ConfigurationMetaNames.PIPELINE_UID : "/v4/pipelines/<PIPELINE-ID>"
         >>> }
         >>> training_details = client.training.run(definition_uid, meta_props=metadata)
         >>> training_uid = client.training.get_uid(training_details)
        """

        WMLResource._chk_and_block_create_update_for_python36(self)
        ##For CP4D, check if either spce or project ID is set
        self._client._check_if_either_is_set()
        Training._validate_type(meta_props, 'meta_props', object, True)
        Training._validate_type(asynchronous, 'asynchronous', bool, True)

        meta_props_str_conv(meta_props)
        self.ConfigurationMetaNames._validate(meta_props)
        training_configuration_metadata = {
            u'training_data_references': meta_props[self.ConfigurationMetaNames.TRAINING_DATA_REFERENCES],
            u'results_reference': meta_props[self.ConfigurationMetaNames.TRAINING_RESULTS_REFERENCE]
        }
        if self.ConfigurationMetaNames.TAGS in meta_props:
            training_configuration_metadata["tags"] = meta_props[self.ConfigurationMetaNames.TAGS]

        # TODO remove when training service starts copying such data on their own

        if self._client.ICP_30 or self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            training_configuration_metadata["name"] = meta_props[self.ConfigurationMetaNames.NAME]
            training_configuration_metadata["description"] = meta_props[self.ConfigurationMetaNames.DESCRIPTION]

            if self.ConfigurationMetaNames.PIPELINE in meta_props:
                training_configuration_metadata["pipeline"] = meta_props[self.ConfigurationMetaNames.PIPELINE]
            if self.ConfigurationMetaNames.EXPERIMENT in meta_props:
                training_configuration_metadata['experiment'] = meta_props[self.ConfigurationMetaNames.EXPERIMENT]
            if self.ConfigurationMetaNames.MODEL_DEFINITION in meta_props:
                training_configuration_metadata['model_definition'] = \
                    meta_props[self.ConfigurationMetaNames.MODEL_DEFINITION]
            if self.ConfigurationMetaNames.SPACE_UID in meta_props:
                training_configuration_metadata["space_id"] = meta_props[self.ConfigurationMetaNames.SPACE_UID]

            if self._client.default_space_id is None and self._client.default_project_id is None:
                raise WMLClientError(
                    "It is mandatory to set the space/project id. Use client.set.default_space(<SPACE_UID>)/client.set.default_project(<PROJECT_UID>) to proceed.")
            else:
                if self._client.default_space_id is not None:
                    training_configuration_metadata['space_id'] = self._client.default_space_id
                elif self._client.default_project_id is not None:
                    training_configuration_metadata['project_id'] = self._client.default_project_id

            if self._client.ICP_30 or self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                if self.ConfigurationMetaNames.FEDERATED_LEARNING in meta_props:
                    training_configuration_metadata['federated_learning'] = \
                        meta_props[self.ConfigurationMetaNames.FEDERATED_LEARNING]
        else:
            if self.ConfigurationMetaNames.PIPELINE_UID in meta_props:
                training_configuration_metadata["pipeline"] = {
                    "href": "/v4/pipelines/"+meta_props[self.ConfigurationMetaNames.PIPELINE_UID]
                }
                if self.ConfigurationMetaNames.PIPELINE_DATA_BINDINGS in meta_props:
                    training_configuration_metadata["pipeline"]["data_bindings"] = meta_props[self.ConfigurationMetaNames.PIPELINE_DATA_BINDINGS]
                if self.ConfigurationMetaNames.PIPELINE_NODE_PARAMETERS in meta_props:
                    training_configuration_metadata["pipeline"]["nodes_parameters"] = meta_props[
                        self.ConfigurationMetaNames.PIPELINE_NODE_PARAMETERS]
                if self.ConfigurationMetaNames.PIPELINE_MODEL_TYPE in meta_props:
                    training_configuration_metadata["pipeline"]["model_type"] = meta_props[
                        self.ConfigurationMetaNames.PIPELINE_MODEL_TYPE]
            if self.ConfigurationMetaNames.EXPERIMENT_UID in meta_props:
                training_configuration_metadata["experiment"] = {
                    "href": "/v4/experiments/" + meta_props[self.ConfigurationMetaNames.EXPERIMENT_UID]
                }

            if self.ConfigurationMetaNames.TRAINING_LIB_UID in meta_props:
                if self._client.CAMS:
                    type_uid = self._check_if_lib_or_def(meta_props[self.ConfigurationMetaNames.TRAINING_LIB_UID])
                    training_configuration_metadata["training_lib"] = {"href" : type_uid}
                else:
                    training_configuration_metadata["training_lib"]["href"] = {"href" : "/v4/libraries/" + meta_props[self.ConfigurationMetaNames.TRAINING_LIB_UID]}

                if self.ConfigurationMetaNames.COMMAND not in meta_props or self.ConfigurationMetaNames.TRAINING_LIB_RUNTIME_UID not in meta_props:
                    raise WMLClientError(u'Invalid input. command, runtime are mandatory parameter for training_lib')
                training_configuration_metadata["training_lib"].update({"command":meta_props[self.ConfigurationMetaNames.COMMAND]})
                training_configuration_metadata["training_lib"].update({"runtime": {"href" : "/v4/runtimes/"+meta_props[self.ConfigurationMetaNames.TRAINING_LIB_RUNTIME_UID]}})
                if self.ConfigurationMetaNames.TRAINING_LIB_MODEL_TYPE in meta_props:
                    training_configuration_metadata["training_lib"].update({"model_type":  meta_props[self.ConfigurationMetaNames.TRAINING_LIB_MODEL_TYPE]})
                if self.ConfigurationMetaNames.COMPUTE in meta_props:
                    training_configuration_metadata["training_lib"].update({"compute": meta_props[self.ConfigurationMetaNames.COMPUTE]})
                if self.ConfigurationMetaNames.TRAINING_LIB_PARAMETERS in meta_props:
                    training_configuration_metadata["training_lib"].update({"parameters": meta_props[self.ConfigurationMetaNames.TRAINING_LIB_PARAMETERS]})

            if self.ConfigurationMetaNames.TRAINING_LIB in meta_props:
                training_configuration_metadata["training_lib"] =  meta_props[self.ConfigurationMetaNames.TRAINING_LIB]
                # for model_definition asset - command, href and runtime are mandatory
                if self._is_model_definition_url(meta_props[self.ConfigurationMetaNames.TRAINING_LIB]['href']) is False:
                    if ('command' not in meta_props[self.ConfigurationMetaNames.TRAINING_LIB].keys() or
                            'runtime' not in meta_props[self.ConfigurationMetaNames.TRAINING_LIB].keys()):
                        raise WMLClientError(u'Invalid input. command, href, runtime are mandatory parameter for training_lib')

            if self.ConfigurationMetaNames.SPACE_UID in meta_props:
                training_configuration_metadata["space"] = {
                    "href": "/v4/spaces/"+meta_props[self.ConfigurationMetaNames.SPACE_UID]
                }
            if self._client.CAMS:
                if self._client.default_space_id is not None:
                    training_configuration_metadata['space'] = {'href': "/v4/spaces/" + self._client.default_space_id}
                elif self._client.default_project_id is not None:
                    training_configuration_metadata['project'] = {'href': "/v2/projects/" + self._client.default_project_id}
                else:
                    raise WMLClientError("It is mandatory to set the space/project id. Use client.set.default_space(<SPACE_UID>)/client.set.default_project(<PROJECT_UID>) to proceed.")

        train_endpoint = self._client.service_instance._href_definitions.get_trainings_href()
        if not self._ICP:
            if self._client.CLOUD_PLATFORM_SPACES:
                params = self._client._params()
                if 'space_id' in params.keys():
                    params.pop('space_id')
                if 'project_id' in params.keys():
                    params.pop('project_id')
                response_train_post = requests.post(train_endpoint, json=training_configuration_metadata,
                                                    params= params,
                                                    headers=self._client._get_headers())
            else:
             response_train_post = requests.post(train_endpoint, json=training_configuration_metadata,
                                                headers=self._client._get_headers())

        else:
            if self._client.ICP_PLATFORM_SPACES:
                params = self._client._params()
                if 'space_id' in params.keys():
                    params.pop('space_id')
                if 'project_id' in params.keys():
                    params.pop('project_id')
                if 'userfs' in params.keys():
                    params.pop('userfs')

                response_train_post = requests.post(train_endpoint,
                                                    json=training_configuration_metadata,
                                                    params=params,
                                                    headers=self._client._get_headers(),
                                                    verify=False)

            elif self._client.ICP_30:
                response_train_post = requests.post(train_endpoint, json=training_configuration_metadata,
                                                    headers=self._client._get_headers(),verify=False)
            else:
                response_train_post = requests.post(train_endpoint, json=training_configuration_metadata,
                                                    headers=self._client._get_headers(), verify=False)

        run_details = self._handle_response(201, u'training', response_train_post)

        try:
            trained_model_guid = self.get_id(run_details)

        except Exception:
            trained_model_guid = self.get_uid(run_details)

        if asynchronous is True:
            return run_details
        else:
            print_text_header_h1(u'Running \'{}\''.format(trained_model_guid))

            status = self.get_status(trained_model_guid)
            state = status[u'state']

            with StatusLogger(state) as status_logger:
                while state not in ['error', 'completed', 'canceled', 'failed']:
                    time.sleep(5)
                    state = self.get_status(trained_model_guid)['state']
                    status_logger.log_state(state)

            if u'completed' in state:
                print(u'\nTraining of \'{}\' finished successfully.'.format(str(trained_model_guid)))
            else:
                print(u'\nTraining of \'{}\' failed with status: \'{}\'.'.format(trained_model_guid, str(status)))

            self._logger.debug(u'Response({}): {}'.format(state, run_details))
            return self.get_details(trained_model_guid)

    def list(self, limit=None, asynchronous=False, get_all=False):
        """
           List stored trainings. If limit is set to None there will be only first 50 records shown.

           **Parameters**

           .. important::
                #. **limit**:  limit number of fetched records at once\n
                   **type**: int\n
                #. **asynchronous**:  if True, it will work as a generator (optional)\n
                   **type**: bool\n
                #. **get_all**:  if True, it will get all entries in 'limited' chunks (optional)\n
                   **type**: bool\n

           **Output**

           .. important::
                This method only prints the list of all trainings in a table format.\n
                **return type**: None\n

           **Example**

            >>> client.training.list()
            >>> training_runs_df = client.training.list(limit=100)
            >>> training_runs_df = client.training.list(limit=100, get_all=True)
            >>> training_runs_df = []
            >>> for entry in client.training.list(limit=100, asynchronous=True, get_all=True):
            >>>    training_runs_df.extend(entry)
        """
        ##For CP4D, check if either spce or project ID is set
        self._client._check_if_either_is_set()
        def preprocess_details(details: dict):
            resources = details[u'resources']
            values = [(m[u'metadata'].get('id', m[u'metadata'].get('guid')), m[u'entity'][u'status'][u'state'],
                       m[u'metadata'][u'created_at']) for m in resources]

            return self._list(values, [u'ID (training)', u'STATE', u'CREATED'], limit=None,
                              default_limit=100000, sort_by=None)

        if asynchronous:
            return (preprocess_details(details) for details in self.get_details(
                limit=limit, asynchronous=asynchronous, get_all=get_all))

        else:
            details = self.get_details(limit=limit, get_all=get_all)
            return preprocess_details(details)

    def list_subtrainings(self, training_uid):
        """
           List the sub-trainings.

           **Parameters**

           .. important::
                #. **training_uid**:  Training ID\n
                   **type**: str\n

           **Output**

           .. important::
                This method only prints the list of all sub-trainings associated with a training in a table format.\n
                **return type**: None\n

           **Example**

            >>> client.training.list_subtrainings()

        """
        ##For CP4D, check if either spce or project ID is set
        if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError("Sub-trainings are no longer available for trainings, please use list_intermediate_models().")

        self._client._check_if_either_is_set()
        details = self.get_details(training_uid)
        verify=True
        if self._client.ICP:
            verify=False
        if "experiment" not in details["entity"]:
            raise WMLClientError("Sub-trainings are available for training created via experiment only.")
        details_parent = requests.get(
            self._wml_credentials['url'] + '/v4/trainings?parent_id=' + training_uid,
            params=self._client._params(),
            headers=self._client._get_headers(),
            verify=verify
        )
        details_json = self._handle_response(200, "Get training details", details_parent)
        resources = details_json["resources"]
        values = [(m[u'metadata'].get('id', m[u'metadata'].get('guid')), m[u'entity'][u'status'][u'state'], m[u'metadata'][u'created_at']) for m in resources]

        self._list(values, [u'ID (sub_training)', u'STATE', u'CREATED'], None, 50)

    def list_intermediate_models(self, training_uid):
        """
           List the intermediate_models.

           **Parameters**

           .. important::
                #. **training_uid**:  Training ID\n
                   **type**: str\n

           **Output**

           .. important::
                This method only prints the list of all intermediate_models associated with an AUTOAI training in a table format.\n
                **return type**: None\n

           .. note::

                This method prints the training logs.
                This method is not supported for IBM Cloud Pak® for Data.

           **Example**

            >>> client.training.list_intermediate_models()

        """
        ##For CP4D, check if either spce or project ID is set
        if self._client.ICP_30 or self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError("This method is not supported for IBM Cloud Pak® for Data. " )

        self._client._check_if_either_is_set()
        details = self.get_details(training_uid)
        #if status is completed then only lists only global_output else display message saying "state value"
        training_state = details[u'entity'][u'status'][u'state']
        if(training_state=='completed'):

            if 'metrics' in details[u'entity'][u'status'] and details[u'entity'][u'status'].get('metrics') is not None:
                metrics_list = details[u'entity'][u'status'][u'metrics']
                new_list=[]
                for ml in metrics_list:
                    # if(ml[u'context'][u'intermediate_model'][u'process']=='global_output'):
                    if 'context' in ml and 'intermediate_model' in ml[u'context']:
                        name = ml[u'context'][u'intermediate_model'].get('name', "")
                        if 'location' in ml[u'context'][u'intermediate_model']:
                            path = ml[u'context'][u'intermediate_model'][u'location'].get('model', "")
                        else:
                            path = ""
                    else:
                        name = ""
                        path = ""

                    accuracy=ml[u'ml_metrics'].get('training_accuracy', "")
                    F1Micro=round(ml[u'ml_metrics'].get('training_f1_micro', 0), 2)
                    F1Macro = round(ml[u'ml_metrics'].get('training_f1_macro', 0), 2)
                    F1Weighted = round(ml[u'ml_metrics'].get('training_f1_weighted', 0), 2)
                    logLoss=round(ml[u'ml_metrics'].get('training_neg_log_loss', 0), 2)
                    PrecisionMicro = round(ml[u'ml_metrics'].get('training_precision_micro', 0), 2)
                    PrecisionWeighted = round(ml[u'ml_metrics'].get('training_precision_weighted', 0), 2)
                    PrecisionMacro = round(ml[u'ml_metrics'].get('training_precision_macro', 0), 2)
                    RecallMacro = round(ml[u'ml_metrics'].get('training_recall_macro', 0), 2)
                    RecallMicro = round(ml[u'ml_metrics'].get('training_recall_micro', 0), 2)
                    RecallWeighted = round(ml[u'ml_metrics'].get('training_recall_weighted', 0), 2)
                    createdAt = details[u'metadata'][u'created_at']
                    new_list.append([name,path,accuracy,F1Micro,F1Macro,F1Weighted,logLoss,PrecisionMicro,PrecisionMacro,PrecisionWeighted,RecallMicro,RecallMacro,RecallWeighted,createdAt])
                    new_list.append([])

                from tabulate import tabulate
                header = [u'NAME', u'PATH', u'Accuracy', u'F1Micro', u'F1Macro', u'F1Weighted', u'LogLoss', u'PrecisionMicro' , u'PrecisionMacro',u'PrecisionWeighted', u'RecallMicro', u'RecallMacro', u'RecallWeighted', u'CreatedAt' ]
                table = tabulate([header] + new_list)

                print(table)
                #self._list(new_list, [u'NAME', u'PATH', u'Accuracy', u'F1Micro', u'F1Macro', u'F1Weighted', u'LogLoss', u'PrecisionMicro' , u'PrecisionMacro',u'PrecisionWeighted', u'RecallMicro', u'RecallMacro', u'RecallWeighted', u'CreatedAt' ], None, 50)
            else:
                print(" There is no intermediate model metrics are available for this training uid. ")
        else:
            self._logger.debug("state is not completed")

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def cancel(self, training_uid, hard_delete=False):
        """
        Cancel a training which is currently running and remove it. This method is also be used to delete metadata
        details of the completed or canceled training run when hard_delete parameter is set to True.

        **Parameters**

        .. important::
            #. **training_uid**:  Training UID\n
               **type**: str\n

            #. **hard_delete**: specify True or False.
               True - To delete the completed or canceled training runs.
               False - To cancel the currently running training run. Default value is False.
               **type**: Boolean\n

        **Output**

        .. important::
            **returns**: status ("SUCCESS" or "FAILED")\n
            **return type**: str\n

        **Example**

         >>> client.training.cancel(training_uid)
        """
        ##For CP4D, check if either spce or project ID is set
        self._client._check_if_either_is_set()
        training_uid = str_type_conv(training_uid)
        Training._validate_type(training_uid, u'training_uid', STR_TYPE, True)
        if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            params = self._client._params()
        else:
            params = None

        if hard_delete is True:
            if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                params.update({'hard_delete': u'true'})
            else:
                params = {}
                params.update({'hard_delete': u'true'})

        if not self._ICP and not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            response_delete = requests.delete(self._client.service_instance._href_definitions.get_training_href(training_uid),
                                          headers=self._client._get_headers(),params=params)
        else:
            response_delete = requests.delete(self._client.service_instance._href_definitions.get_training_href(training_uid),
                                              headers=self._client._get_headers(),params=params,verify=False)

        if response_delete.status_code == 400 and \
           response_delete.text is not None and 'Job already completed with state' in response_delete.text:
            print("Job is not running currently. Please use 'hard_delete=True' parameter to force delete"
                  " completed or canceled training runs.")
            return "SUCCESS"
        else:
            return self._handle_response(204, u'trained model deletion', response_delete, False)

    def _COS_logs(self, run_uid,on_start=lambda: {}):
        on_start()
        run_details = self.get_details(run_uid)
        if 'connection' in run_details["entity"]["results_reference"] and run_details["entity"]["results_reference"].get("connection") is not None:
            endpoint_url = run_details["entity"]["results_reference"]["connection"]["endpoint_url"]
            aws_access_key = run_details["entity"]["results_reference"]["connection"]["access_key_id"]
            aws_secret = run_details["entity"]["results_reference"]["connection"]["secret_access_key"]
            bucket = run_details["entity"]["results_reference"]["location"]["bucket"]
            # try:
            #     run_details["entity"]["training_results_reference"]["location"]["model_location"]
            # except:
            #     raise WMLClientError("The training-run has not started. Error - " + run_details["entity"]["status"]["error"]["errors"][0]["message"])

            if (bucket == ""):
                bucket = run_details["entity"]["results_reference"]["target"]["bucket"]
            import ibm_boto3

            client_cos = ibm_boto3.client(service_name='s3', aws_access_key_id=aws_access_key,
                                          aws_secret_access_key=aws_secret,
                                          endpoint_url=endpoint_url)

            try:
                if self._client.CLOUD_PLATFORM_SPACES:
                    logs = run_details["entity"].get("results_reference").get("location").get("logs")
                    if logs is None:
                        print(" There is no logs details for this Training run, hence no logs.")
                        return

                    key = logs + "/learner-1/training-log.txt"

                else:
                    try:
                        key = "data/" + run_details["metadata"].get("id", run_details["metadata"].get('guid')) + "/pipeline-model.json"

                        obj = client_cos.get_object(Bucket=bucket, Key=key)
                        pipeline_model = json.loads((obj['Body'].read().decode('utf-8')))

                    except ibm_boto3.exceptions.ibm_botocore.client.ClientError as ex:
                        if ex.response['Error']['Code'] == 'NoSuchKey':
                            print(" Error - There is no training logs are found for the given training run id")
                            return
                         #   print("ERROR - Cannot find pipeline_model.json in the bucket "+ run_uid)
                        else:
                            print(ex)
                            return
                    if pipeline_model is not None:
                        key = pipeline_model["pipelines"][0]["nodes"][0]["parameters"]["model_id"] + "/learner-1/training-log.txt"
                    else:
                        print(" Error - Cannot find the any logs for the given training run id")
                obj = client_cos.get_object(Bucket=bucket, Key=key)
                print(obj['Body'].read().decode('utf-8'))
            except ibm_boto3.exceptions.ibm_botocore.client.ClientError as ex:

                if ex.response['Error']['Code'] == 'NoSuchKey':
                    print("ERROR - Cannot find training-log.txt in the bucket")
                else:
                    print(ex)
                    print("ERROR - Cannot get the training run log in the bucket")
        else:
            print(" There is no connection details for this Training run, hence no logs.")


    def _COS_metrics(self, run_uid,on_start=lambda: {}):
        on_start()
        run_details = self.get_details(run_uid)
        endpoint_url = run_details["entity"]["results_reference"]["connection"]["endpoint_url"]
        aws_access_key = run_details["entity"]["results_reference"]["connection"]["access_key_id"]
        aws_secret = run_details["entity"]["results_reference"]["connection"]["secret_access_key"]
        bucket = run_details["entity"]["results_reference"]["location"]["bucket"]
        # try:
        #     run_details["entity"]["training_results_reference"]["location"]["model_location"]
        # except:
        #     raise WMLClientError("The training-run has not started. Error - " + run_details["entity"]["status"]["error"]["errors"][0]["message"])

        if (bucket == ""):
            bucket = run_details["entity"]["results_reference"]["target"]["bucket"]
        import ibm_boto3

        client_cos = ibm_boto3.client(service_name='s3', aws_access_key_id=aws_access_key,
                                      aws_secret_access_key=aws_secret,
                                      endpoint_url=endpoint_url)

        try:
            if self._client.CLOUD_PLATFORM_SPACES:
                logs = run_details["entity"].get("results_reference").get("location").get("logs")
                if logs is None:
                    print(" Metric log location details for this Training run is not available.")
                    return
                key = logs + "/learner-1/evaluation-metrics.txt"
            else:
                try:
                    key = run_details["metadata"].get("id", run_details["metadata"].get('guid')) + "/pipeline-model.json"

                    obj = client_cos.get_object(Bucket=bucket, Key=key)

                    pipeline_model = json.loads((obj['Body'].read().decode('utf-8')))
                except ibm_boto3.exceptions.ibm_botocore.client.ClientError as ex:

                    if ex.response['Error']['Code'] == 'NoSuchKey':
                        print("ERROR - Cannot find pipeline_model.json in the bucket for training id "+ run_uid)
                        print("There is no training logs are found for the given training run id")
                        return
                    else:
                        print(ex)
                        return
                key = pipeline_model["pipelines"][0]["nodes"][0]["parameters"].get["model_id"] + "/learner-1/evaluation-metrics.txt"

            obj = client_cos.get_object(Bucket=bucket, Key=key)
            print(obj['Body'].read().decode('utf-8'))

        except ibm_boto3.exceptions.ibm_botocore.client.ClientError as ex:
            if ex.response['Error']['Code'] == 'NoSuchKey':
                print("ERROR - Cannot find evaluation-metrics.txt in the bucket")
            else:
                print(ex)
                print("ERROR - Cannot get the location of evaluation-metrics.txt details in the bucket")

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def monitor_logs(self, training_uid):
        """
            Monitor the logs of a training created.

            **Parameters**

            .. important::
                #. **training_uid**:  Training UID\n
                   **type**: str\n

            **Output**

            .. important::

                **returns**: None\n
                **return type**: None\n

            .. note::

                This method prints the training logs.
                This method is not supported for IBM Cloud Pak® for Data.

            **Example**

             >>> client.training.monitor_logs(training_uid)

        """

        if self._client.ICP_30 or self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError("Metrics logs are not supported. This method is not supported for IBM Cloud Pak® for Data. ")

        training_uid = str_type_conv(training_uid)
        Training._validate_type(training_uid, u'training_uid', STR_TYPE, True)

        self._simple_monitor_logs(training_uid, lambda: print_text_header_h1(u'Log monitor started for training run: ' + str(training_uid)))

        print_text_header_h2('Log monitor done.')

    def _simple_monitor_logs(self, training_uid, on_start=lambda: {}):
        try:
            run_details = self.get_details(training_uid)
        except ApiRequestFailure as ex:
            if "404" in str(ex.args[1]):
                print("Could not find the training run details for the given training run id. ")
                return
            else:
                raise ex

        status = run_details["entity"]["status"]["state"]

        if (status == "completed" or status == "error" or status == "failed" or status == "canceled"):
            self._COS_logs(training_uid,
                           lambda: print_text_header_h1(u'Log monitor started for training run: ' + str(training_uid)))
        else:
            if not self._ICP:
                if self._client.CLOUD_PLATFORM_SPACES:
                    ws_param = self._client._params()
                    if 'project_id' in ws_param.keys():
                        proj_id = ws_param.get('project_id')
                        monitor_endpoint = self._wml_credentials[u'url'].replace(u'https',
                                                                                 u'wss') + u'/ml/v4/trainings/' + training_uid + "?project_id=" + proj_id
                    else:
                        space_id = ws_param.get('space_id')
                        monitor_endpoint = self._wml_credentials[u'url'].replace(u'https',
                                                                                 u'wss') + u'/ml/v4/trainings/' + training_uid + "?space_id=" + space_id
                else:
                    monitor_endpoint = self._wml_credentials[u'url'].replace(u'https',
                                                                     u'wss') + u'/v4/trainings/' + training_uid
            else:
                monitor_endpoint = self._wml_credentials[u'url'].replace(u'https',
                                                                         u'wss') + u'/v4/trainings/' + training_uid
            websocket = WebSocket(monitor_endpoint)
            try:
                websocket.add_header(bytes("Authorization", "utf-8"), bytes("Bearer " + self._client.service_instance._get_token(), "utf-8"))
            except:
                websocket.add_header(bytes("Authorization"), bytes("bearer " + self._client.service_instance._get_token()))

            on_start()

            for event in websocket:

                if event.name == u'text':
                    text = json.loads(event.text)
                    entity = text[u'entity']
                    if 'status' in entity:
                      if 'message' in entity['status']:
                        message = entity['status']['message']
                        if len(message) > 0:
                          print(message)

            websocket.close()

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def monitor_metrics(self, training_uid):
        """
            Monitor the metrics of a training created.

            **Parameters**

            .. important::
                #. **training_uid**:  Training UID\n
                   **type**: str\n

            **Output**

            .. important::

                **returns**: None\n
                **return type**: None\n

            .. note::

                This method prints the training metrics.
                This method is not supported for IBM Cloud Pak® for Data.

            **Example**

             >>> client.training.monitor_metrics(training_uid)
        """
        if self._client.ICP_30 or self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError("Metrics monitoring is not supported for IBM Cloud Pak® for Data")

        training_uid = str_type_conv(training_uid)
        Training._validate_type(training_uid, u'training_uid', STR_TYPE, True)
        try:
            run_details = self.get_details(training_uid)
        except ApiRequestFailure as ex:
            if "404" in str(ex.args[1]):
                print("Could not find the training run details for the given training run id. ")
                return
            else:
                raise ex
        status = run_details["entity"]["status"]["state"]

        if (status == "completed" or status == "error" or status == "failed" or status == "canceled"):
            self._COS_metrics(training_uid,
                           lambda: print_text_header_h1(u'Log monitor started for training run: ' + str(training_uid)))
        else:
            if not self._ICP:
                if self._client.CLOUD_PLATFORM_SPACES:
                    ws_param = self._client._params()
                    if 'project_id' in ws_param.keys():
                        proj_id = ws_param.get('project_id')
                        monitor_endpoint = self._wml_credentials[u'url'].replace(u'https',
                                                                                 u'wss') + u'/ml/v4/trainings/' + training_uid + "?project_id=" + proj_id
                    else:
                        space_id = ws_param.get('space_id')
                        monitor_endpoint = self._wml_credentials[u'url'].replace(u'https',
                                                                                 u'wss') + u'/ml/v4/trainings/' + training_uid + "?space_id=" + space_id

                else:
                    monitor_endpoint = self._wml_credentials[u'url'].replace(u'https',
                                                                     u'wss') + u'/v4/trainings/' + training_uid
            else:
                monitor_endpoint = self._wml_credentials[u'url'].replace(u'https',
                                                                         u'wss') + u'/v4/trainings/' + training_uid
            websocket = WebSocket(monitor_endpoint)
            try:
                websocket.add_header(bytes("Authorization", "utf-8"), bytes("Bearer " + self._client.service_instance._get_token(), "utf-8"))
            except:
                websocket.add_header(bytes("Authorization"), bytes("bearer " + self._client.service_instance._get_token()))

            print_text_header_h1('Metric monitor started for training run: ' + str(training_uid))

            for event in websocket:
                if event.name == u'text':
                    text = json.loads(event.text)
                    entity = text[u'entity']
                    if 'status' in entity:
                        status = entity[u'status']
                        if u'metrics' in status:
                            metrics = status[u'metrics']
                            if len(metrics) > 0:
                             metric = metrics[0]
                             print(metric)

            websocket.close()

            print_text_header_h2('Metric monitor done.')

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_metrics(self, training_uid):
        """
                Get metrics.

                **Parameters**

                .. important::
                    #. **training_uid**:  training UID\n
                       **type**: str\n

                **Output**

                .. important::
                    **returns**: Metrics of a training run\n
                    **return type**: list of dict\n

                **Example**

                 >>> training_status = client.training.get_metrics(training_uid)

        """

        training_uid = str_type_conv(training_uid)
        Training._validate_type(training_uid, u'training_uid', STR_TYPE, True)
        status = self.get_status(training_uid)
        if 'metrics' in status:
            return status['metrics']
        else:
            details = self.get_details(training_uid)
            if 'metrics' in details:
                return details['metrics']
            else:
                raise WMLClientError("No metrics details are available for the given training_uid")

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def _get_latest_metrics(self, training_uid):
        """
             Get latest metrics values.

             :param training_uid: ID of trained model
             :type training_uid: {0}

             :returns: metric values
             :rtype: list of dicts

             A way you might use me is:

             >>> client.training.get_latest_metrics(training_uid)
         """
        training_uid = str_type_conv(training_uid)
        Training._validate_type(training_uid, u'training_uid', STR_TYPE, True)

        status = self.get_status(training_uid)
        metrics = status.get('metrics', [])
        latest_metrics = []

        if len(metrics) > 0:
            grouped_metrics = group_metrics(metrics)

            for key, value in grouped_metrics.items():
                sorted_value = sorted(value, key=lambda k: k['iteration'])

            latest_metrics.append(sorted_value[-1])

        return latest_metrics


import unittest

import logging
from ibm_watson_machine_learning.tests.CP4D_35.preparation_and_cleaning import *

class TestFunction(unittest.TestCase):
    runtime_uid = None
    deployment_uid = None
    function_id = None
    function_id2 = None
    scoring_url = None
    logger = logging.getLogger(__name__)

    @classmethod
    def setUpClass(self):
        TestFunction.logger.info("Service Instance: setting up credentials")

        self.wml_credentials = get_wml_credentials()
        self.client = get_client()


        # self.space_id = get_project_id()
        # self.client.set.default_project(self.space_id)


        self.space_name = str(uuid.uuid4())

        metadata = {
                     self.client.spaces.ConfigurationMetaNames.NAME: 'client_space_' + self.space_name,
                     self.client.spaces.ConfigurationMetaNames.DESCRIPTION: self.space_name + ' description'
        }

        self.space = self.client.spaces.store(meta_props=metadata)

        TestFunction.space_id = self.client.spaces.get_id(self.space)
        print("space_id: ", TestFunction.space_id)

        self.client.set.default_space(TestFunction.space_id)
        # self.client.set.default_space('db7b719a-c095-4bb9-94e8-29b93b51beb2')
        # self.client.set.default_space('b88a8678-c653-4a2e-88b7-870e9655f752')

        import time
        time.sleep(30)

    def test_01_create_function(self):

        self.client.repository.FunctionMetaNames.show()
        # self.client.software_specifications.list()
        sw_spec_id = '0cdb0f1e-5376-4f4d-92dd-da3b69aa9bda'
        print('sw_spec_id: ', sw_spec_id)

        function_meta_props = {
            self.client.repository.FunctionMetaNames.NAME: 'sample_function_with_sw',
            self.client.repository.FunctionMetaNames.SOFTWARE_SPEC_ID: sw_spec_id #,
            # self.client.repository.FunctionMetaNames.SAMPLE_SCORING_INPUT: {
            #     "input_data": [
            #       {
            #         "fields": [
            #           "name",
            #           "age",
            #           "occupation"
            #         ],
            #         "values": [
            #           [
            #             "john",
            #             23,
            #             "student"
            #           ],
            #           [
            #             "paul",
            #             33,
            #             "engineer"
            #           ]
            #         ]
            #       }
            #     ]
            #   } #,
            # self.client.repository.FunctionMetaNames.INPUT_DATA_SCHEMAS: [{
            #     "id": "id",
            #     "name": "input_data_schema",
            #     "fields": [
            #         {
            #             "name": "x",
            #             "type": "double",
            #             "nullable": False
            #             # "metadata": {"measure":"default", "modeling_role":"none"}
            #         },
            #         {
            #             "name": "y",
            #             "type": "double",
            #             "nullable": False
            #             # "metadata": {"measure":"default", "modeling_role":"none"}
            #         }
            #     ]
            # }],
            # self.client.repository.FunctionMetaNames.OUTPUT_DATA_SCHEMAS: [{
            #     "id": "id",
            #     "name": "output_data_schema",
            #     "fields": [
            #         {
            #             "name": "multiplication",
            #             "type": "double",
            #             "nullable": False
            #             # "metadata": {"measure":"default", "modeling_role":"none"}
            #         }
            #     ]
            # }]
        }

        # def test_update():
        #     def score(payload):
        #         return {'predictions': [{'fields': [], 'values': [['working']]}]}
        #
        #     return score
        #
        def score(payload):
            payload = payload['input_data'][0]
            values = [[row[0] * row[1]] for row in payload['values']]
            return {'predictions': [{'fields': ['multiplication'], 'values': values}]}

        # def score_generator():
        #     import urllib3, requests, json, time, os
        #     import subprocess
        #     # Define scoring function
        #     def callModel(payload_scoring):
        #         ## user code
        #         import sys
        #         import os
        #         import sklearn
        #         import numpy as np
        #         skl_ver = sklearn.__version__
        #         np_ver = np.__version__
        #         v4_scoring_response = {
        #             'predictions': [{'fields': ['sklearn', 'np'],
        #                              'values': [[skl_ver, np_ver]]
        #                              }]
        #         }
        #         return v4_scoring_response
        #
        #     def score(input):
        #         """AI function example.
        #         Example:
        #             {
        #                 "input_data": [{
        #                         "id": "1",
        #                         "fields": [
        #                             "name",
        #                             "age",
        #                             "occupation"
        #                         ],
        #                         "values": [
        #                             [
        #                                 "john",
        #                                 23,
        #                                 "student"
        #                             ],
        #                             [
        #                                 "paul",
        #                                 33,
        #                                 "engineer"
        #                             ]
        #                         ]
        #                     }
        #                 ]
        #             }
        #         """
        #         # Score using the pre-defined model
        #         prediction = callModel(input);
        #         return prediction
        #
        #     return score

        # function_details = self.client.repository.store_function(score_generator, function_meta_props)
        function_details = self.client.repository.store_function(score, function_meta_props)
        # function_details = self.client.repository.store_function(test_update, function_meta_props)

        print(function_details)

        TestFunction.function_id = self.client.repository.get_function_id(function_details)
        function_url = self.client.repository.get_function_href(function_details)
        TestFunction.logger.info("function ID:" + str(TestFunction.function_id))
        TestFunction.logger.info("function URL:" + str(function_url))
        self.assertIsNotNone(TestFunction.function_id)
        self.assertIsNotNone(function_url)

    def test_02_create_function2(self):

        self.client.repository.FunctionMetaNames.show()
        sw_spec_id = self.client.software_specifications.get_id_by_name("ai-function_0.1-py3.6")

        print(sw_spec_id)

        function_meta_props = {
            self.client.repository.FunctionMetaNames.NAME: 'sample_function2_with_sw',
            self.client.repository.FunctionMetaNames.SOFTWARE_SPEC_ID: sw_spec_id,
            self.client.repository.FunctionMetaNames.TYPE: 'python'
        }

        function_details2 = self.client.repository.store_function(function='artifacts/score.py.gz',
                                                                  meta_props=function_meta_props)
        # function_details2 = self.client.repository.store_function(function='/Users/mitun_vb/test_function.gz',
        #                                                           meta_props=function_meta_props)

        print(function_details2)

        TestFunction.function_id2 = self.client.repository.get_function_id(function_details2)
        function_url2 = self.client.repository.get_function_href(function_details2)
        TestFunction.logger.info("function ID:" + str(TestFunction.function_id2))
        TestFunction.logger.info("function URL:" + str(function_url2))
        self.assertIsNotNone(TestFunction.function_id2)
        self.assertIsNotNone(function_url2)

    def test_03_get_details(self):

        details = self.client.repository.get_function_details(TestFunction.function_id)
        print(details)
        self.assertTrue('sample_function' in str(details))

        details2 = self.client.repository.get_function_details(TestFunction.function_id2)
        print(details2)
        self.assertTrue('sample_function2' in str(details2))

    def test_04_list(self):
        # self.client.repository.list()
        self.client.repository.list_functions()

    def test_05_download_ai_function_content(self):
        try:
            os.remove('test_function.gz')
        except:
            pass
        self.client.repository.download(TestFunction.function_id, filename='test_function.gz')

        try:
            os.remove('test_function.gz')
        except:
            pass

        try:
            os.remove('test_function1.gz')
        except:
            pass
        self.client.repository.download(TestFunction.function_id2, filename='test_function1.gz')

        try:
            os.remove('test_function1.gz')
        except:
            pass

    def test_06_update_function(self):
        function_update_meta_props = {
            self.client.repository.FunctionMetaNames.NAME: 'updated_function_name'
        }

        # def test_update():
        #     def score(payload):
        #         return {'predictions': [{'fields': [], 'values': [['working']]}]}
        #
        #     return score

        details = self.client.repository.update_function(TestFunction.function_id,
                                                         function_update_meta_props,
                                                         # update_function=test_update)
                                                         update_function='artifacts/score_rev1.py.gz')
        self.assertFalse('sample_function' in json.dumps(details))
        self.assertTrue('updated_function_name' in json.dumps(details))

        try:
            os.remove('test_function_update.gz')
        except:
            pass
        self.client.repository.download(TestFunction.function_id, filename='test_function_update.gz')

        try:
            os.remove('test_function_update.gz')
        except:
            pass

    def test_07_revisions(self):
        revision = self.client.repository.create_function_revision(TestFunction.function_id)
        print('revision: ', revision)
        self.assertTrue(revision[u'metadata'][u'rev'] == '1')

        revision_details = self.client.repository.get_function_revision_details(TestFunction.function_id, 1)
        print('revision details: ', revision_details)
        self.assertTrue(revision[u'metadata'][u'rev'] == '1')

        self.client.repository.list_functions_revisions(TestFunction.function_id)

        try:
            os.remove('test_function_revision.gz')
        except:
            pass

        # Download revision 1 attachment
        self.client.repository.download(TestFunction.function_id,
                                        filename='test_function_revision.gz',
                                        rev_uid=1)
        try:
            os.remove('test_function_revision.gz')
        except:
            pass


    def test_08_create_deployment(self):
        deploy_meta = {
                self.client.deployments.ConfigurationMetaNames.NAME: "deployment_Function",
                self.client.deployments.ConfigurationMetaNames.DESCRIPTION: "deployment_description",
                self.client.deployments.ConfigurationMetaNames.ONLINE: {}
            }

        TestFunction.logger.info("Create deployment")
        deployment = self.client.deployments.create(artifact_uid=TestFunction.function_id, meta_props=deploy_meta)
        TestFunction.logger.debug("deployment: " + str(deployment))
        TestFunction.scoring_url = self.client.deployments.get_scoring_href(deployment)
        TestFunction.logger.debug("Scoring href: {}".format(TestFunction.scoring_url))
        TestFunction.deployment_id = self.client.deployments.get_id(deployment)
        TestFunction.logger.debug("Deployment uid: {}".format(TestFunction.deployment_id))
        self.client.deployments.list()
        self.assertTrue("deployment_Function" in str(deployment))

    def test_09_update_deployment(self):
        patch_meta = {
            self.client.deployments.ConfigurationMetaNames.DESCRIPTION: "deployment_Updated_Function_Description",
        }
        self.client.deployments.update(TestFunction.deployment_id, patch_meta)

    def test_10_get_deployment_details(self):
        TestFunction.logger.info("Get deployment details")
        deployment_details = self.client.deployments.get_details()
        TestFunction.logger.debug("Deployment details: {}".format(deployment_details))
        self.assertTrue('deployment_Function' in str(deployment_details))

    def test_11_score(self):
        scoring_payload = {
            self.client.deployments.ScoringMetaNames.INPUT_DATA: [{
                "fields": ["multiplication"],
                "values": [[2.0, 2.0], [99.0, 99.0]]
            }
            ]
        }
        predictions = self.client.deployments.score(TestFunction.deployment_id, scoring_payload)
        print("Predictions: {}".format(predictions))
        self.assertTrue("values" in str(predictions))

    def test_12_delete_deployment(self):
        TestFunction.logger.info("Delete deployment")
        self.client.deployments.delete(TestFunction.deployment_id)

    def test_13_delete_function(self):
        TestFunction.logger.info("Delete function")
        self.client.repository.delete(TestFunction.function_id)
        self.client.repository.delete(TestFunction.function_id2)

    def test_14_delete_space(self):
        self.client.spaces.delete(TestFunction.space_id)

if __name__ == '__main__':
    unittest.main()



import unittest
import time

import logging
from ibm_watson_machine_learning.tests.CP4D_35.preparation_and_cleaning import *
import time
import requests

class TestPkgExtn(unittest.TestCase):

    logger = logging.getLogger(__name__)

    @classmethod
    def setUpClass(self):
        TestPkgExtn.logger.info("Service Instance: setting up credentials")

        self.wml_credentials = get_wml_credentials()
        self.client = get_client()

        # self.space_id = get_project_id()
        # self.client.set.default_project(self.space_id)


        self.space_name = str(uuid.uuid4())

        metadata = {
                     self.client.spaces.ConfigurationMetaNames.NAME: 'client_space_' + self.space_name,
                     self.client.spaces.ConfigurationMetaNames.DESCRIPTION: self.space_name + ' description'
        }

        self.space = self.client.spaces.store(meta_props=metadata)
        TestPkgExtn.space_id = self.client.spaces.get_id(self.space)
        # TestRshinyApp.space_id = '9cf73498-72d6-42eb-a6f4-424616de9f45'
        # TestPkgExtn.space_id = '590a241b-756b-4364-a8ae-380b3e6607f7'
        print('space_id: ', TestPkgExtn.space_id)

        print(self.client.spaces.get_details(TestPkgExtn.space_id))
        time.sleep(20)
        print(self.client.spaces.get_details(TestPkgExtn.space_id))
        self.client.set.default_space(TestPkgExtn.space_id)


    def test_00_create_pkg_extn(self):

        meta_prop_pkg_extn = {
            self.client.package_extensions.ConfigurationMetaNames.NAME: "Pkg extension for conda",
            self.client.package_extensions.ConfigurationMetaNames.DESCRIPTION: "Pkg extension for conda",
            self.client.package_extensions.ConfigurationMetaNames.TYPE: "conda_yml",
        }

        # print(self.client.spaces.get_details(TestPkgExtn.space_id))
        pkg_extn_details = self.client.package_extensions.store(meta_props=meta_prop_pkg_extn,
                                                                file_path="artifacts/conda_regex.yaml")

        TestPkgExtn.pkg_extn_id = self.client.package_extensions.get_id(pkg_extn_details)
        pkg_extn_url = self.client.package_extensions.get_href(pkg_extn_details)
        TestPkgExtn.logger.info("sw spec ID:" + str(TestPkgExtn.pkg_extn_id))
        TestPkgExtn.logger.info("w spec URL:" + str(pkg_extn_url))

        print("pkg_extn_id: {}; pkg_extn_url: {}".format(TestPkgExtn.pkg_extn_id, pkg_extn_url))

        self.assertIsNotNone(TestPkgExtn.pkg_extn_id)
        self.assertIsNotNone(pkg_extn_url)

    def test_01_get_details(self):

        details = self.client.package_extensions.get_details(TestPkgExtn.pkg_extn_id)
        print(details)
        self.assertTrue("Pkg extension for conda" in str(details))

        href = self.wml_credentials[u'url'] + details[u'entity'][u'package_extension'][u'href']

        try:
            os.remove('test.yaml')
        except:
            pass

        response = requests.get(href, verify=False)

        print('href: ', href)
        print('response: ', response)

        with open('test.yaml', 'wb') as f:
            f.write(response.content)

        try:
            os.remove('test.yaml')
        except:
            pass


    def test_02_list_package_extensions(self):
        self.client.package_extensions.list()

    def test_03_create_sw_spec(self):

        self.client.software_specifications.ConfigurationMetaNames.show()
        # base_sw_spec_id = self.client.software_specifications.get_id_by_name("ai-function_0.1-py3.6")
        base_sw_spec_id = self.client.software_specifications.get_id_by_name("default_py3.7")

        meta_prop_sw_spec = {
            self.client.software_specifications.ConfigurationMetaNames.NAME: "sw_spec test" + str(uuid.uuid4()),
            self.client.software_specifications.ConfigurationMetaNames.DESCRIPTION: "Software specification test",
            self.client.software_specifications.ConfigurationMetaNames.BASE_SOFTWARE_SPECIFICATION:
                {"guid": base_sw_spec_id}
        }

        sw_spec_details = self.client.software_specifications.store(meta_props=meta_prop_sw_spec)

        print(sw_spec_details)

        TestPkgExtn.sw_spec_id = self.client.software_specifications.get_id(sw_spec_details)
        sw_spec_url = self.client.software_specifications.get_href(sw_spec_details)
        TestPkgExtn.logger.info("sw spec ID:" + str(TestPkgExtn.sw_spec_id))
        TestPkgExtn.logger.info("sw spec URL:" + str(sw_spec_url))

        print("sw_spec_id: {}; sw_spec_url: {}".format(TestPkgExtn.sw_spec_id, sw_spec_url))

        self.assertIsNotNone(TestPkgExtn.sw_spec_id)
        self.assertIsNotNone(sw_spec_url)

    def test_04_add_pkg_extn_to_sw_spec(self):

        self.client.software_specifications.add_package_extension(TestPkgExtn.sw_spec_id,
                                                                  TestPkgExtn.pkg_extn_id)

        ss_asset_details = self.client.software_specifications.get_details(TestPkgExtn.sw_spec_id)
        print('Package extensions',
              print(ss_asset_details['entity']['software_specification']['package_extensions']))

    def test_05_get_details(self):

        details = self.client.software_specifications.get_details(TestPkgExtn.sw_spec_id)
        print(details)
        self.assertTrue("sw_spec test" in str(details))
        self.assertTrue("Pkg extension for conda" in str(details))

    def test_06_del_pkg_extn_to_sw_spec(self):

        self.client.software_specifications.delete_package_extension(TestPkgExtn.sw_spec_id,
                                                                     TestPkgExtn.pkg_extn_id)

    def test_07_get_details(self):

        details = self.client.software_specifications.get_details(TestPkgExtn.sw_spec_id)
        print(details)
        self.assertTrue("sw_spec test" in str(details))
        self.assertTrue("Pkg extension for conda" not in str(details))

    def test_08_list(self):
        self.client.software_specifications.list()

    def test_09_delete_pkg_extn(self):
        TestPkgExtn.logger.info("Delete deployment")
        self.client.package_extensions.delete(TestPkgExtn.pkg_extn_id)

    def test_10_delete_sw_spec(self):
        TestPkgExtn.logger.info("Delete deployment")
        self.client.software_specifications.delete(TestPkgExtn.sw_spec_id)

    def test_11_delete_space(self):
        TestPkgExtn.logger.info("Delete space")
        self.client.spaces.delete(TestPkgExtn.space_id)

if __name__ == '__main__':
    unittest.main()



import unittest
from sklearn.pipeline import Pipeline
from pprint import pprint
import pandas as pd
import traceback
from os import environ

from ibm_watson_machine_learning import APIClient

from ibm_watson_machine_learning.experiment import AutoAI
from ibm_watson_machine_learning.deployment import WebService, Batch
from ibm_watson_machine_learning.workspace import WorkSpace

from ibm_watson_machine_learning.helpers.connections import S3Connection, S3Location, DataConnection, FSLocation

from ibm_watson_machine_learning.experiment.autoai.optimizers import RemoteAutoPipelines

from ibm_watson_machine_learning.utils.autoai.errors import NotExistingCOSResource, ForecastingUnsupportedOperation

from ibm_watson_machine_learning.tests.utils import (get_wml_credentials, get_cos_credentials, bucket_exists,
                                                     create_bucket, get_space_id, is_cp4d)
from ibm_watson_machine_learning.tests.utils.cleanup import space_cleanup, delete_model_deployment

from ibm_watson_machine_learning.utils.autoai.enums import PositiveLabelClass, TShirtSize, Transformers


class TestAutoAIRemote(unittest.TestCase):
    """
    The test can be run on CLOUD, WMLS and CPD (not tested)
    The test covers:
    - COS set-up (if run on Cloud): checking if bucket exists for the cos instance, if not new bucket is create
    - Saving data `/bank.cdv` to COS/data assets
    - downloading training data from cos/data assets
    - downloading all generated pipelines to lale pipeline
    - deployment with lale pipeline
    - deployment deletion
    """
    wml_client: 'APIClient' = None
    experiment: 'AutoAI' = None
    remote_auto_pipelines: 'RemoteAutoPipelines' = None
    wml_credentials = None
    cos_credentials = None
    pipeline_opt: 'RemoteAutoPipelines' = None
    service: 'WebService' = None
    service_batch: 'Batch' = None

    data_location = './autoai/data/Twitter_volume_AMZN_multivariate.csv'

    trained_pipeline_details = None
    run_id = None

    data_connection = None
    results_connection = None

    train_data = None

    bucket_name = environ.get('BUCKET_NAME', "wml-autoaitests-qa")
    pod_version = environ.get('KB_VERSION', None)
    space_name = environ.get('SPACE_NAME', 'regression_tests_sdk_space')

    cos_endpoint = "https://s3.us-south.cloud-object-storage.appdomain.cloud"
    data_cos_path = 'data/Twitter_volume_AMZN.csv'
    cos_resource_instance_id = None

    results_cos_path = 'results_wml_autoai'

    pipeline: 'Pipeline' = None
    lale_pipeline = None
    deployed_pipeline = None
    hyperopt_pipelines = None
    new_pipeline = None
    new_sklearn_pipeline = None
    X_values = None
    y_values = None

    OPTIMIZER_NAME = 'COVID csv wml_autoai forecasting test'
    DEPLOYMENT_NAME = "COVID AutoAI Deployment tests"

    project_id = None
    space_id = None

    asset_id = None

    @classmethod
    def setUpClass(cls) -> None:
        """
        Load WML credentials from config.ini file based on ENV variable.
        """
        cls.wml_credentials = get_wml_credentials()
        cls.wml_client = APIClient(wml_credentials=cls.wml_credentials.copy())

        if not cls.wml_client.ICP:
            cls.cos_credentials = get_cos_credentials()
            cls.cos_endpoint = cls.cos_credentials.get('endpoint_url')
            cls.cos_resource_instance_id = cls.cos_credentials.get('resource_instance_id')

        # cls.project_id = cls.wml_credentials.get('project_id')


    def test_00a_space_cleanup(self):
        space_cleanup(self.wml_client,
                      get_space_id(self.wml_client, self.space_name,
                                   cos_resource_instance_id=self.cos_resource_instance_id),
                      days_old=7)
        TestAutoAIRemote.space_id = get_space_id(self.wml_client, self.space_name,
                                     cos_resource_instance_id=self.cos_resource_instance_id)

        self.wml_client.set.default_space(self.space_id)

    def test_00b_prepare_COS_instance(self):
        if self.wml_client.ICP:
            self.skipTest("Prepare COS is available only for Cloud")
        import ibm_boto3
        cos_resource = ibm_boto3.resource(
            service_name="s3",
            endpoint_url=self.cos_endpoint,
            aws_access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
            aws_secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key']
        )
        # Prepare bucket
        if not bucket_exists(cos_resource, self.bucket_name):
            TestAutoAIRemote.bucket_name = create_bucket(cos_resource, self.bucket_name)

            self.assertIsNotNone(TestAutoAIRemote.bucket_name)
            self.assertTrue(bucket_exists(cos_resource, TestAutoAIRemote.bucket_name))

        print(f"Using COS bucket: {TestAutoAIRemote.bucket_name}")

    def test_01_initialize_AutoAI_experiment__pass_credentials__object_initialized(self):
        TestAutoAIRemote.experiment = AutoAI(wml_credentials=self.wml_credentials.copy(),
                                             space_id=self.space_id)

        self.assertIsInstance(self.experiment, AutoAI, msg="Experiment is not of type AutoAI.")

    def test_02_save_remote_data_and_DataConnection_setup(self):
        if self.wml_client.ICP:
            TestAutoAIRemote.data_connection = DataConnection(
                            location=FSLocation(path=self.data_location))
            TestAutoAIRemote.results_connection = None

        else: #for cloud and COS
            TestAutoAIRemote.data_connection = DataConnection(
                connection=S3Connection(endpoint_url=self.cos_endpoint,
                                        access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
                                        secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key']),
                location=S3Location(bucket=self.bucket_name,
                                    path=self.data_cos_path)
            )
            TestAutoAIRemote.results_connection = DataConnection(
                connection=S3Connection(endpoint_url=self.cos_endpoint,
                                        access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
                                        secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key']),
                location=S3Location(bucket=self.bucket_name,
                                    path=self.results_cos_path)
            )
            TestAutoAIRemote.data_connection.write(data=self.data_location, remote_name=self.data_cos_path)

        self.assertIsNotNone(obj=TestAutoAIRemote.data_connection)

    def test_02a_read_saved_remote_data_before_fit(self):
        data = self.data_connection.read()
        print(f"Data sample")
        print(data.head())
        self.assertGreater(len(data), 0)

    def test_03_initialize_optimizer(self):
        TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(
            name=self.OPTIMIZER_NAME,
            desc='test description',
            prediction_type=self.experiment.PredictionType.FORECASTING,
            prediction_columns=['value1', 'value2'],
            timestamp_column_name='timestamp',
            backtest_num=4,
            lookback_window=5,
            forecast_window=2,
            holdout_size=0.05,
            max_number_of_estimators=1,
            include_only_estimators=[AutoAI.ForecastingAlgorithms.ENSEMBLER],
            t_shirt_size=TShirtSize.S if self.wml_client.ICP else TShirtSize.L,
            autoai_pod_version=self.pod_version,
            #csv_separator=','

        )

        self.assertIsInstance(self.remote_auto_pipelines, RemoteAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

    def test_04_get_configuration_parameters_of_remote_auto_pipeline(self):
        parameters = self.remote_auto_pipelines.get_params()
        print(parameters)
        self.assertIsInstance(parameters, dict, msg='Config parameters are not a dictionary instance.')

    def test_05_fit_run_training_of_auto_ai_in_wml(self):
        TestAutoAIRemote.trained_pipeline_details = self.remote_auto_pipelines.fit(
            training_data_reference=[self.data_connection],
            training_results_reference=self.results_connection,
            background_mode=False)

        TestAutoAIRemote.run_id = self.trained_pipeline_details['metadata']['id']

        self.assertIsNotNone(self.data_connection.auto_pipeline_params,
                             msg='DataConnection auto_pipeline_params was not updated.')

    def test_05a_fit_error_when_cos_file_not_exists(self):
        if self.wml_client.ICP:
            self.skipTest("Prepare COS is available only for Cloud")

        with self.assertRaises(NotExistingCOSResource, msg="Training should raise error when non existing COS bucket given"):
            self.remote_auto_pipelines.fit(
                training_data_reference=[
                    DataConnection(
                    connection=S3Connection(endpoint_url=self.cos_endpoint,
                                            access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
                                            secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key']),
                    location=S3Location(bucket="not_existing",
                                        path=self.data_cos_path)
                )],
                training_results_reference=self.results_connection,
                background_mode=False)

        with self.assertRaises(NotExistingCOSResource, msg="Training should raise error when non existing COS file given"):
            self.remote_auto_pipelines.fit(
                training_data_reference=[
                    DataConnection(
                        connection=S3Connection(endpoint_url=self.cos_endpoint,
                                                access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
                                                secret_access_key=self.cos_credentials['cos_hmac_keys'][
                                                    'secret_access_key']),
                        location=S3Location(bucket=self.bucket_name,
                                            path="not_existing")
                    )],
                training_results_reference=self.results_connection,
                background_mode=False)

    def test_06_get_train_data(self):
        TestAutoAIRemote.train_data = self.remote_auto_pipelines.get_data_connections()[0].read()

        print("train data sample:")
        print(self.train_data.head())
        self.assertGreater(len(self.train_data), 0)

        TestAutoAIRemote.X_values = self.train_data.drop(['value1', 'value2'], axis=1)[:10].values
        print(self.train_data.head())
        TestAutoAIRemote.y_values = self.train_data[['value1', 'value2']][:10]

    def test_07_get_run_status(self):
        status = self.remote_auto_pipelines.get_run_status()
        self.assertEqual(status, "completed", msg="AutoAI run didn't finished successfully. Status: {}".format(status))

    def test_08_get_run_details(self):
        parameters = self.remote_auto_pipelines.get_run_details()
        import json
        print(json.dumps(self.wml_client.training.get_details(training_uid=parameters['metadata']['id']), indent=4))
        print(parameters)
        self.assertIsNotNone(parameters)

    def test_08b_get_metrics(self):
        self.wml_client.training.get_metrics(self.run_id)

    def test_09_predict_using_fitted_pipeline(self):
        with self.assertRaises(ForecastingUnsupportedOperation, msg="Operation should be unsupported"):
            self.remote_auto_pipelines.predict(X=self.X_values[:5])

    def test_10_summary_listing_all_pipelines_from_wml(self):
        pipelines_details = self.remote_auto_pipelines.summary()
        print(pipelines_details.to_string())

    def test_11__get_data_connections__return_a_list_with_data_connections_with_optimizer_params(self):
        data_connections = self.remote_auto_pipelines.get_data_connections()
        self.assertIsInstance(data_connections, list, msg="There should be a list container returned")
        self.assertIsInstance(data_connections[0], DataConnection,
                              msg="There should be a DataConnection object returned")

    def test_12_get_pipeline_params_specific_pipeline_parameters(self):
        pipeline_params = self.remote_auto_pipelines.get_pipeline_details(pipeline_name='Pipeline_1')
        print(pipeline_params)

    def test_13__get_pipeline_params__fetch_best_pipeline_parameters__parameters_fetched_as_dict(self):
        best_pipeline_params = self.remote_auto_pipelines.get_pipeline_details()
        print(best_pipeline_params)

    ####
    # LALE #
    ########

    def test_14__get_pipeline__load_lale_pipeline__pipeline_loaded(self):
        with self.assertRaises(ForecastingUnsupportedOperation, msg="Operation should be unsupported"):
            self.remote_auto_pipelines.get_pipeline(pipeline_name='Pipeline_4')

    def test_15_get_all_pipelines_as_lale(self):
        summary = self.remote_auto_pipelines.summary()
        print(summary.to_string())
        for pipeline_name in summary.reset_index()['Pipeline Name']:
            print(f"Getting pipeline: {pipeline_name}")
            with self.assertRaises(ForecastingUnsupportedOperation, msg="Operation should be unsupported"):
                self.remote_auto_pipelines.get_pipeline(pipeline_name=pipeline_name)

    #################################
    #      DEPLOYMENT SECTION       #
    #################################

    def test_21_deployment_setup_and_preparation(self):
        TestAutoAIRemote.service = WebService(source_wml_credentials=self.wml_credentials.copy(),
                                              source_space_id=self.space_id,
                                              target_wml_credentials=self.wml_credentials,
                                              target_space_id=self.space_id)
        self.wml_client.set.default_space(self.space_id)
        delete_model_deployment(self.wml_client, deployment_name=self.DEPLOYMENT_NAME)

        self.assertIsInstance(self.service, WebService, msg="Deployment is not of WebService type.")
        self.assertIsInstance(self.service._source_workspace, WorkSpace, msg="Workspace set incorrectly.")
        self.assertEqual(self.service.id, None, msg="Deployment ID initialized incorrectly")
        self.assertEqual(self.service.name, None, msg="Deployment name initialized incorrectly")

    def test_22__deploy__deploy_best_computed_pipeline_from_autoai_on_wml(self):
        best_pipeline = self.remote_auto_pipelines.summary()._series['Enhancements'].keys()[0]
        print('Deploying', best_pipeline)
        self.service.create(
            experiment_run_id=self.remote_auto_pipelines._engine._current_run_id,
            model=best_pipeline,
            deployment_name=self.DEPLOYMENT_NAME)

        self.assertIsNotNone(self.service.id, msg="Online Deployment creation - missing id")
        self.assertIsNotNone(self.service.name, msg="Online Deployment creation - name not set")
        self.assertIsNotNone(self.service.scoring_url,
                             msg="Online Deployment creation - mscoring url  missing")

    def test_23_score_deployed_model(self):
        predictions = self.service.score(payload=pd.DataFrame({'Cases' : []}))
        print(predictions)
        self.assertIsNotNone(predictions)

    def test_24_list_deployments(self):
        self.service.list()
        params = self.service.get_params()
        print(params)
        self.assertIsNotNone(params)

    def test_25_delete_deployment(self):
        print("Delete current deployment: {}".format(self.service.deployment_id))
        self.service.delete()
        self.wml_client.set.default_space(self.space_id) if not self.wml_client.default_space_id else None
        self.wml_client.repository.delete(self.service.asset_id)
        self.assertEqual(self.service.id, None, msg="Deployment ID deleted incorrectly")
        self.assertEqual(self.service.name, None, msg="Deployment name deleted incorrectly")
        self.assertEqual(self.service.scoring_url, None,
                         msg="Deployment scoring_url deleted incorrectly")

    #########################
    #  Batch deployment
    #########################

    def test_30_batch_deployment_setup_and_preparation(self):
        TestAutoAIRemote.service_batch = Batch(source_wml_credentials=self.wml_credentials.copy(),
                                               source_space_id=self.space_id,
                                               target_wml_credentials=self.wml_credentials,
                                               target_space_id=self.space_id)

        self.assertIsInstance(self.service_batch, Batch, msg="Deployment is not of Batch type.")
        self.assertIsInstance(self.service_batch._source_workspace, WorkSpace, msg="Workspace set incorrectly.")
        self.assertEqual(self.service_batch.id, None, msg="Deployment ID initialized incorrectly")
        self.assertEqual(self.service_batch.name, None, msg="Deployment name initialized incorrectly")

    def test_32__deploy__batch_deploy_best_computed_pipeline_from_autoai_on_wml(self):
        self.service_batch.create(
            experiment_run_id=self.remote_auto_pipelines._engine._current_run_id,
            model="Pipeline_2",
            deployment_name=self.DEPLOYMENT_NAME + ' BATCH')

        self.assertIsNotNone(self.service_batch.id, msg="Batch Deployment creation - missing id")
        self.assertIsNotNone(self.service_batch.id, msg="Batch Deployment creation - name not set")
        self.assertIsNotNone(self.service_batch.asset_id,
                          msg="Batch Deployment creation - model (asset) id missing, incorrect model storing")

    def test_33_score_batch_deployed_model(self):
        scoring_params = self.service_batch.run_job(
            payload=self.train_data.drop(columns=['day']),
            background_mode=False)
        print(scoring_params)
        print(self.service_batch.get_job_result(scoring_params['metadata']['id']))
        self.assertIsNotNone(scoring_params)
        self.assertIn('predictions', str(scoring_params).lower())

    def test_34_list_batch_deployments(self):
        deployments = self.service_batch.list()
        print(deployments)
        self.assertIn('created_at', str(deployments).lower())
        self.assertIn('status', str(deployments).lower())

        params = self.service_batch.get_params()
        print(params)
        self.assertIsNotNone(params)

    def test_35_delete_deployment_batch(self):
        print("Delete current deployment: {}".format(self.service_batch.deployment_id))
        self.service_batch.delete()
        self.wml_client.set.default_space(self.space_id) if not self.wml_client.default_space_id else None
        self.wml_client.repository.delete(self.service_batch.asset_id)
        self.assertEqual(self.service_batch.id, None, msg="Deployment ID deleted incorrectly")
        self.assertEqual(self.service_batch.name, None, msg="Deployment name deleted incorrectly")
        self.assertEqual(self.service_batch.scoring_url, None,
                         msg="Deployment scoring_url deleted incorrectly")
if __name__ == '__main__':
    unittest.main()



import unittest

from ibm_watson_machine_learning.helpers import DataConnection, S3Connection, S3Location
from ibm_watson_machine_learning.tests.utils import get_cos_credentials, get_wml_credentials
from ibm_watson_machine_learning.experiment import AutoAI
from ibm_watson_machine_learning.experiment.autoai.optimizers.local_auto_pipelines import LocalAutoPipelines
import pandas as pd

"""
Format of COS credentials:

cos_credentials = {
                  "auth_endpoint": "https://iam.cloud.ibm.com/identity/token",
                  "endpoint_url": "https://s3.us.cloud-object-storage.appdomain.cloud",
                  "apikey": "...", 
                  "cos_hmac_keys": {
                    "access_key_id": "...",
                    "secret_access_key": "..."
                  }
                  }

endpoint_url should be appropriate per bucket location, see on UI.
"""


class TestOBMOutput(unittest.TestCase):
    cos_credentials = None
    bucket = 'test-donotdelete-pr-7w72prz2oteazs'
    path = 'bank.csv'
    pipeline_model_hmac = None
    pipeline_model_auth_endpoint = None
    metadata_hmac = None
    metadata_auth_endpoint = None
    pipeline_name = 'Pipeline_1'
    training_status = 'auto_ml/5e04fc63-906f-46d1-af47-c4d76ba5c42a/wml_data/c8ed6196-6187-4e80-997f-378a1b9b24e2/training-status.json'
    model_location = 'auto_ml/5e04fc63-906f-46d1-af47-c4d76ba5c42a/wml_data/c8ed6196-6187-4e80-997f-378a1b9b24e2/data/kb/global_output/'
    run_id = 'c8ed6196-6187-4e80-997f-378a1b9b24e2'

    local_optimizer_hmac: 'LocalAutoPipelines' = None
    local_optimizer: 'LocalAutoPipelines' = None

    data_location = './autoai/data/bank.csv'
    data = None
    X = None
    y = None

    @classmethod
    def setUp(cls) -> None:
        cls.wml_credentials = get_wml_credentials()
        cls.cos_credentials = get_cos_credentials()
        cls.data = pd.read_csv(cls.data_location)
        cls.X = cls.data.drop(['y'], axis=1)
        cls.y = cls.data['y']

    def test_01_get_preprocessed_data_connection_with_holdout_split_notebook_COS_version(self):
        print("HMAC: Initializing DataConnections with COS credentials and paths...")

        training_data_reference = [DataConnection(
            connection=S3Connection(
                access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
                secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key'],
                endpoint_url=self.cos_credentials['endpoint_url']
            ),
            location=S3Location(
                bucket=self.bucket,
                path=self.path,
            )
        )]

        training_result_reference = DataConnection(
            connection=S3Connection(
                access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
                secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key'],
                endpoint_url=self.cos_credentials['endpoint_url']
            ),
            location=S3Location(
                bucket=self.bucket,
                path=self.path,
                model_location=self.model_location,
                training_status=self.training_status
            )
        )

        TestOBMOutput.metadata_hmac = dict(
            training_data_reference=training_data_reference,
            training_result_reference=training_result_reference,
            prediction_type='regression',
            prediction_column='next_purchase',
            test_size=0.1,
            scoring='roc_auc',
            max_number_of_estimators=1,
        )

        print("HMAC: Initializing AutoAI local scenario with metadata...")

        TestOBMOutput.local_optimizer_hmac = AutoAI().runs.get_optimizer(
            metadata=self.metadata_hmac
        )

        training_df, holdout_df = self.local_optimizer_hmac.get_preprocessed_data_connection().read(
            with_holdout_split=True)

        print(f"Training data frame: {training_df}")
        print(f"Holdout data frame: {holdout_df}")

    def test_02_get_preprocessed_data_connection_with_holdout_historical_scenario(self):
        experiment = AutoAI(self.wml_credentials)
        optimizer = experiment.runs.get_optimizer(run_id=self.run_id)

        training_df, holdout_df = optimizer.get_preprocessed_data_connection().read(
            with_holdout_split=True)

        print(f"Training data frame: {training_df}")
        print(f"Holdout data frame: {holdout_df}")


if __name__ == '__main__':
    unittest.main()



"""
IBM Confidential
OCO Source Materials
5737-H76, 5725-W78, 5900-A1R
(c) Copyright IBM Corp. 2020 All Rights Reserved.
The source code for this program is not published or otherwise divested of its trade secrets,
irrespective of what has been deposited with the U.S. Copyright Office.
"""
import abc
import os
from pathlib import Path
import logging

import ibmfl.envs as fl_envs

logger = logging.getLogger(__name__)


class FLModel(abc.ABC):
    """
    Base class for all FLModels. This includes supervised and unsupervised ones.
    """

    def __init__(self, model_name, model_spec, **kwargs):
        """
        Initializes an `FLModel` object

        :param model_name: String describing the model e.g., keras_cnn
        :type model_name: `str`
        :param model_spec: Specification of the the model
        :type model_spec: `dict`
        :param kwargs: Dictionary of model-specific arguments.
        :type kwargs: `dict`
        """
        self.model_name = model_name
        # reserve to specify the model type by FLModel class
        self.model_type = None
        self.model_spec = model_spec if model_spec else {}
        self.use_gpu_for_training = False
        if 'info' in kwargs and kwargs['info'] is not None:
            gpu_config = kwargs['info'].get('gpu')
            # load GPU setting
            if gpu_config:
                self.use_gpu_for_training = True
                if 'selection' in gpu_config and gpu_config.get('selection') == 'auto':
                    self.num_gpus, self.gpu_ids = self.get_gpu_config_auto(gpu_config)
                else:
                    self.num_gpus, self.gpu_ids = self.get_gpu_config_manual(gpu_config)

                os.environ['CUDA_VISIBLE_DEVICES'] = self.gpu_ids

    def get_custom_metrics_pre(self):
        """
        Return pre-training metrics to log via metrics manager.

        :return: Dictionary of custom metrics to pass through straight to metrics manager.
        :rtype: `dict`
        """

        return {}

    def get_custom_metrics_post(self):
        """
        Return post-training metrics to log via metrics manager.

        :return: Dictionary of custom metrics to pass through straight to metrics manager.
        :rtype: `dict`
        """

        return {}

    def get_train_result(self):
        raise NotImplementedError

    @abc.abstractmethod
    def fit_model(self, train_data, **kwargs):
        """
        Fits current model with provided training data.

        :param train_data: Training data.
        :type train_data: Data structure containing training data, \
        varied based on model types.
        :param kwargs: Dictionary of model-specific arguments for fitting, \
        e.g., hyperparameters for local training, information provided \
        by aggregator, etc.
        :type kwargs: `dict`
        :return: None
        """
        raise NotImplementedError

    @abc.abstractmethod
    def update_model(self, model_update, **kwargs):
        """
        Updates model using provided `model_update`. Additional arguments
        specific to the model can be added through `**kwargs`

        :param model_update: Model with update. This is specific to each model \
        type e.g., `ModelUpdateSGD`. The specific type should be checked by \
        the corresponding FLModel class.
        :type model_update: `ModelUpdate`
        :param kwargs: Dictionary of model-specific arguments.
        :type kwargs: `dict`
        :return: None
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_model_update(self):
        """
        Generate a `ModelUpdate` object specific to the FLModel being trained.
        This object will be shared with other parties.

        :return: Model Update. Object specific to model being trained.
        :rtype: `ModelUpdate`
        """
        raise NotImplementedError

    @abc.abstractmethod
    def predict(self, x, **kwargs):
        """
        Perform prediction for a batch of inputs.

        :param x: Samples with shape as expected by the model.
        :type x: Data structure as expected by the model
        :param kwargs: Dictionary of model-specific arguments.
        :type kwargs: `dict`

        :return: Predictions
        :rtype: Data structure the same as the type defines labels \
        in testing data.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def evaluate(self, test_dataset, batch_size=128, **kwargs):
        """
        Evaluates model given the test dataset.
        Multiple evaluation metrics are returned in a dictionary

        :param test_dataset: Provided test dataset to evalute the model.
        :type test_dataset: `tuple` of `np.ndarray` or data generator.
        :param batch_size: batch_size: Size of batches.
        :type batch_size: `int`
        :param kwargs: Dictionary of model-specific arguments.
        :type kwargs: `dict`
        :return: Dictionary with all evaluation metrics provided by specific \
        implementation.
        :rtype: `dict`
        """
        raise NotImplementedError

    @abc.abstractmethod
    def save_model(self, filename, path=None):
        """
        Save a model to file in the format specific to the backend framework.

        :param filename: Name of the file where to store the model.
        :type filename: `str`
        :param path: Path of the folder where to store the model. If no path \
        is specified, the model will be stored in \
                     the default data location of the library `DATA_PATH`.
        :type path: `str`
        :return: filename
        """
        raise NotImplementedError

    @abc.abstractmethod
    def load_model(self, filename):
        """
        Load model from provided filename

        :param filename: Name of the file where to store the model.
        :type filename: `str`
        :param path: Path of the folder where to store the model. If no path is specified, \
        the model will be stored in the default data location of the library `DATA_PATH`.
        :type path: `str`
        :return: model
        """
        raise NotImplementedError

    def is_fitted(self):
        """
        Return a boolean value indicating if the model is fitted or not.

        :return: res
        :rtype: `bool`
        """
        raise NotImplementedError

    @staticmethod
    def get_model_absolute_path(filename):
        """Construct absolute path using MODEL_DIR env variable

        :param filename: Name of the file
        :type filename: `str`
        :return: absolute_path; constructed absolute path using model_dir
        :rtype: `str``

        """
        if "MODEL_DIR" in os.environ:
            model_path = Path(os.environ["MODEL_DIR"])
        else:
            model_path = Path(fl_envs.model_directory)

        model_path.mkdir(parents=True, exist_ok=True)
        absolute_path = model_path.joinpath(filename)
        return str(absolute_path)

    def get_loss(self, dataset):
        """
        Return the resulting loss computed based on the provided dataset.

        :param dataset: Provided dataset, a tuple given in the form \
        (x_test, y_test) or a datagenerator
        :type dataset: `np.ndarray`
        :return: The resulting loss.
        :rtype: `float`
        """
        raise NotImplementedError


    def get_gpu_config_auto(self, gpu_config):
        """
        Extract num of gpus and gpu ids automatically using gpu util.
        :param gpu_config: Provided gpu config
        :type gpu_config:  `dict`
        """
        num_gpus = 1
        gpu_ids = ''

        import GPUtil
        _gpu_ids = GPUtil.getFirstAvailable(order='memory')
        num_gpus = gpu_config.get('num_gpus') if 'num_gpus' in gpu_config else len(_gpu_ids)
        gpu_ids = str(gpuID[0:gpu_num - 1]).replace('[', '').replace(']', '')
        return num_gpus, gpu_ids

    def get_gpu_config_manual(self, gpu_config):
        """
        Extract num of gpus and gpu ids from user provided gpu configuration.
        :param gpu_config: Provided gpu config
        :type gpu_config:  `dict`
        """
        num_gpus = 1
        gpu_ids = ''
        if not gpu_config:
            logger.error("No gpu configuration provided for this job.")

        if 'num_gpus' not in gpu_config and 'gpu_id' not in gpu_config:
            logger.error("No valid gpu configuration provided for this job.")

        if 'gpu_id' in gpu_config:
            gpu_ids = gpu_config.get('gpu_ids')
            num_gpus = len(gpu_ids.split(','))
        
        elif 'num_gpus' in gpu_config:
            if not isinstance(gpu_config.get('num_gpus'), int):
                logger.error(
                    'Provided number of gpu should be of type int. '
                    'Instead it is of type ' +
                    str(type(gpu_config.get('num_gpus'))) +
                    'Set number of gpu to default value 1.')
                self.num_gpus = 1
            else:
                self.num_gpus = gpu_config.get('num_gpus')
        

        return num_gpus, gpu_ids



import unittest
from ibm_watson_machine_learning.helpers.connections import DataConnection, FSLocation
from ibm_watson_machine_learning.tests.utils import is_cp4d
from ibm_watson_machine_learning.tests.autoai.abstract_tests_classes import (
    AbstractTestAutoAIRemote)


@unittest.skip("Not ready, training does not support connection_asset")
@unittest.skipIf(not (is_cp4d()), "Supported only on CP4D")
class TestAutoAIRemote(AbstractTestAutoAIRemote, unittest.TestCase):
    """
    The test can be run only on CPD
    The test covers:
    - File system connection set-up
    - downloading training data from filesystem
    - downloading all generated pipelines to lale pipeline
    - deployment with lale pipeline
    - deployment deletion
    Connection used in test:
     - input: File system.
     - output: [None] - default connection depending on environment.
    """

    def test_02_DataConnection_setup(self):
        TestAutoAIRemote.data_connection = DataConnection(
            location=FSLocation(path=self.data_location)
        )
        TestAutoAIRemote.results_connection = None

        self.assertIsNotNone(obj=TestAutoAIRemote.data_connection)

    def test_02a_read_saved_remote_data_before_fit(self):
        TestAutoAIRemote.data = self.data_connection.read(csv_separator=self.custom_separator)
        print("Data sample:")
        print(self.data.head())
        self.assertGreater(len(self.data), 0)


if __name__ == '__main__':
    unittest.main()



import unittest
import logging
from ibm_watson_machine_learning.wml_client_error import ApiRequestFailure
from ibm_watson_machine_learning.tests.CP4D_35.preparation_and_cleaning import *
from configparser import RawConfigParser


class TestWMLClientWithSPSS_MSSQL(unittest.TestCase):
    deployment_id = None
    model_id = None
    scoring_url = None
    scoring_id = None
    logger = logging.getLogger(__name__)

    connection_sqlserver_asset_uid = None
    sqlserver_connected_data_asset_url1 = None
    sqlserver_connected_data_asset_url2 = None
    sqlserver_connected_data_asset_url3 = None
    sqlserver_connected_data_id1 = None
    sqlserver_connected_data_id2 = None
    sqlserver_connected_data_id3 = None

    config = RawConfigParser()

    @classmethod
    def setUpClass(self):
        TestWMLClientWithSPSS_MSSQL.logger.info("Service Instance: setting up credentials")

        self.wml_credentials = get_wml_credentials()
        self.client = get_client()

        self.space_name = str(uuid.uuid4())

        metadata = {
            self.client.spaces.ConfigurationMetaNames.NAME: 'client_space_' + self.space_name,
            self.client.spaces.ConfigurationMetaNames.DESCRIPTION: self.space_name + ' description'
        }

        self.space = self.client.spaces.store(meta_props=metadata)

        TestWMLClientWithSPSS_MSSQL.space_id = self.client.spaces.get_id(self.space)
        print("space_id: ", TestWMLClientWithSPSS_MSSQL.space_id)
        self.client.set.default_space(TestWMLClientWithSPSS_MSSQL.space_id)
        self.model_path = os.path.join(os.getcwd(), 'artifacts', 'spss-churn-sqlserver.str')

    def test_01_publish_local_model_in_repository(self):
        TestWMLClientWithSPSS_MSSQL.logger.info("Saving trained model in repo ...")
        TestWMLClientWithSPSS_MSSQL.logger.debug("Model path: {}".format(self.model_path))

        self.client.repository.ModelMetaNames.show()

        sw_spec_id = self.client.software_specifications.get_id_by_name("spss-modeler_18.1")

        model_meta_props = {self.client.repository.ModelMetaNames.NAME: "spss-churn-model",
                       self.client.repository.ModelMetaNames.TYPE: "spss-modeler_18.1",
                       self.client.repository.ModelMetaNames.SOFTWARE_SPEC_UID: sw_spec_id
                            }
        published_model = self.client.repository.store_model(model=self.model_path, meta_props=model_meta_props)
        print(published_model)
        TestWMLClientWithSPSS_MSSQL.model_id = self.client.repository.get_model_id(published_model)
        TestWMLClientWithSPSS_MSSQL.logger.info("Published model ID:" + str(TestWMLClientWithSPSS_MSSQL.model_id))
        self.assertIsNotNone(TestWMLClientWithSPSS_MSSQL.model_id)

    def test_02_get_model_details(self):
        TestWMLClientWithSPSS_MSSQL.logger.info("Get model details")
        self.assertIsNotNone(self.client.repository.get_model_details(TestWMLClientWithSPSS_MSSQL.model_id))
        
    def test_03_create_connection_mssql__asset(self):
        import json
        self.client.connections.ConfigurationMetaNames.show()
        self.client.connections.list_datasource_types()
        sqlserver_data_source_type_id = self.client.connections.get_datasource_type_uid_by_name('sqlserver')
        print(sqlserver_data_source_type_id)
        conn_properties = json.loads(config.get(environment, 'sqlserver'))
        meta_props_connection = {
            self.client.connections.ConfigurationMetaNames.NAME: "Sample Connection for sqlserver from Python client",
            self.client.connections.ConfigurationMetaNames.DATASOURCE_TYPE: sqlserver_data_source_type_id,
            self.client.connections.ConfigurationMetaNames.DESCRIPTION: "Test connection object using sqlserver instance",
            self.client.connections.ConfigurationMetaNames.PROPERTIES: conn_properties
        }

        conn_details = self.client.connections.create(meta_props_connection)

        TestWMLClientWithSPSS_MSSQL.connection_sqlserver_asset_uid = self.client.connections.get_uid(conn_details)
        TestWMLClientWithSPSS_MSSQL.logger.info("Connection asset ID:" + str(TestWMLClientWithSPSS_MSSQL.connection_sqlserver_asset_uid))
        self.assertIsNotNone(TestWMLClientWithSPSS_MSSQL.connection_sqlserver_asset_uid)

    def test_04_create_connected_sqlserver_input_data(self):
        asset_meta_props = {
            self.client.data_assets.ConfigurationMetaNames.NAME: "sqlserver",
            self.client.data_assets.ConfigurationMetaNames.CONNECTION_ID: TestWMLClientWithSPSS_MSSQL.connection_sqlserver_asset_uid,
            self.client.data_assets.ConfigurationMetaNames.DESCRIPTION: "sqlserver",
            self.client.data_assets.ConfigurationMetaNames.DATA_CONTENT_NAME: "/CHURN/CHURN3"
            # self.client.data_assets.ConfigurationMetaNames.DATA_CONTENT_NAME: "/b4p0kaxbycnhqvfyt3xj/CHURN"
        }
        asset_details = self.client.data_assets.store(asset_meta_props)
        print(asset_details)
        TestWMLClientWithSPSS_MSSQL.sqlserver_connected_data_asset_url1 = self.client.data_assets.get_href(asset_details)
        TestWMLClientWithSPSS_MSSQL.sqlserver_connected_data_id1 = self.client.data_assets.get_uid(asset_details)
        self.assertIsNotNone(TestWMLClientWithSPSS_MSSQL.sqlserver_connected_data_id1)
        self.assertIsNotNone(TestWMLClientWithSPSS_MSSQL.sqlserver_connected_data_asset_url1)

    # def test_05_create_connected_sqlserver_input_data2(self):
    #     asset_meta_props = {
    #         self.client.data_assets.ConfigurationMetaNames.NAME: "sqlserver",
    #         self.client.data_assets.ConfigurationMetaNames.CONNECTION_ID: TestWMLClientWithSPSS_MSSQL.connection_sqlserver_asset_uid,
    #         self.client.data_assets.ConfigurationMetaNames.DESCRIPTION: "sqlserver",
    #         self.client.data_assets.ConfigurationMetaNames.DATA_CONTENT_NAME: "/CHURN/CHURN4"
    #         # self.client.data_assets.ConfigurationMetaNames.DATA_CONTENT_NAME: "/b4p0kaxbycnhqvfyt3xj/CHURN"
    #     }
    #     asset_details = self.client.data_assets.store(asset_meta_props)
    #     print(asset_details)
    #     TestWMLClientWithSPSS_MSSQL.sqlserver_connected_data_asset_url2 = self.client.data_assets.get_href(asset_details)
    #     TestWMLClientWithSPSS_MSSQL.sqlserver_connected_data_id2 = self.client.data_assets.get_uid(asset_details)
    #     self.assertIsNotNone(TestWMLClientWithSPSS_MSSQL.sqlserver_connected_data_id2)
    #     self.assertIsNotNone(TestWMLClientWithSPSS_MSSQL.sqlserver_connected_data_asset_url2)

    def test_06_create_connected_sqlserver_output_data(self):
        asset_meta_props = {
            self.client.data_assets.ConfigurationMetaNames.NAME: "sqlserver",
            self.client.data_assets.ConfigurationMetaNames.CONNECTION_ID: TestWMLClientWithSPSS_MSSQL.connection_sqlserver_asset_uid,
            self.client.data_assets.ConfigurationMetaNames.DESCRIPTION: "sqlserver",
            self.client.data_assets.ConfigurationMetaNames.DATA_CONTENT_NAME: "/CHURN/CHURN6"
            # self.client.data_assets.ConfigurationMetaNames.DATA_CONTENT_NAME: "/b4p0kaxbycnhqvfyt3xj/CHURN"
        }
        asset_details = self.client.data_assets.store(asset_meta_props)
        print(asset_details)
        TestWMLClientWithSPSS_MSSQL.sqlserver_connected_data_asset_url3 = self.client.data_assets.get_href(asset_details)
        TestWMLClientWithSPSS_MSSQL.sqlserver_connected_data_id3 = self.client.data_assets.get_uid(asset_details)
        self.assertIsNotNone(TestWMLClientWithSPSS_MSSQL.sqlserver_connected_data_id3)
        self.assertIsNotNone(TestWMLClientWithSPSS_MSSQL.sqlserver_connected_data_asset_url3)

    def test_07_create_deployment(self):
        TestWMLClientWithSPSS_MSSQL.logger.info("Create deployment for sqlserver connected data asset")
        deployment_details = self.client.deployments.create(TestWMLClientWithSPSS_MSSQL.model_id,
                     meta_props={self.client.deployments.ConfigurationMetaNames.NAME: "Test spss deployment with sqlserver",
                                 self.client.deployments.ConfigurationMetaNames.BATCH:{},
                                 self.client.deployments.ConfigurationMetaNames.HARDWARE_SPEC: {
                                                                                                     "name": "S",
                                                                                                     "num_nodes": 1
                                                                                               }})
        
        print(deployment_details)

        TestWMLClientWithSPSS_MSSQL.logger.debug("Deployment details: {}".format(deployment_details))

        TestWMLClientWithSPSS_MSSQL.deployment_id = self.client.deployments.get_id(deployment_details)
        TestWMLClientWithSPSS_MSSQL.logger.debug("Deployment uid: {}".format(TestWMLClientWithSPSS_MSSQL.deployment_id))

    def test_08_create_job(self):
        TestWMLClientWithSPSS_MSSQL.logger.info("Create job details")
        job_payload_ref = {
            self.client.deployments.ScoringMetaNames.INPUT_DATA_REFERENCES: [{
                "name": "sqlserver_ref_input",
                "type": "data_asset",
                "connection": {},
                "location": {
                    "href": TestWMLClientWithSPSS_MSSQL.sqlserver_connected_data_asset_url1
                }
            }],
            self.client.deployments.ScoringMetaNames.OUTPUT_DATA_REFERENCE: {
                "type": "data_asset",
                "connection": {},
                "location": {
                    "href": TestWMLClientWithSPSS_MSSQL.sqlserver_connected_data_asset_url3
                }
            }
        }

        TestWMLClientWithSPSS_MSSQL.job_details = self.client.deployments.create_job(TestWMLClientWithSPSS_MSSQL.deployment_id,
                                                                               meta_props=job_payload_ref)
        print(TestWMLClientWithSPSS_MSSQL.job_details)
        TestWMLClientWithSPSS_MSSQL.job_id = self.client.deployments.get_job_uid(TestWMLClientWithSPSS_MSSQL.job_details)

    def test_09_get_job_status(self):
        import time

        start_time = time.time()
        diff_time = start_time - start_time
        while True and diff_time < 10 * 60:
            time.sleep(3)
            response = self.client.deployments.get_job_status(TestWMLClientWithSPSS_MSSQL.job_id)
            if response['state'] == 'completed' or response['state'] == 'error' or response['state'] == 'failed':
                break
            diff_time = time.time() - start_time

        print(response)
        self.assertIsNotNone(response)
       # self.assertTrue(response['state'] == 'completed')

    def test_10_list_jobs(self):
        self.client.deployments.list_jobs()

    def test_11_delete_job(self):
        self.client.deployments.delete_job(TestWMLClientWithSPSS_MSSQL.job_id)

    def test_12_delete_deployment(self):
        TestWMLClientWithSPSS_MSSQL.logger.info("Delete deployment")
        self.client.deployments.delete(TestWMLClientWithSPSS_MSSQL.deployment_id)
        
    def test_13_delete_connection_asset(self):
        TestWMLClientWithSPSS_MSSQL.logger.info("Delete connection and connected data")
        self.client.data_assets.delete(TestWMLClientWithSPSS_MSSQL.sqlserver_connected_data_id1)
        self.client.data_assets.delete(TestWMLClientWithSPSS_MSSQL.sqlserver_connected_data_id3)
        self.client.connections.delete(TestWMLClientWithSPSS_MSSQL.connection_sqlserver_asset_uid)

    def test_14_delete_model(self):
        TestWMLClientWithSPSS_MSSQL.logger.info("Delete model")
        self.client.repository.delete(TestWMLClientWithSPSS_MSSQL.model_id)
        # self.client.spaces.delete(TestWMLClientWithSPSS_MSSQL.space_id)

    def test_15_delete_space(self):
        self.client.spaces.delete(TestWMLClientWithSPSS_MSSQL.space_id)

if __name__ == '__main__':
    unittest.main()



"""
IBM Confidential
OCO Source Materials
5737-H76, 5725-W78, 5900-A1R
(c) Copyright IBM Corp. 2020 All Rights Reserved.
The source code for this program is not published or otherwise divested of its trade secrets,
irrespective of what has been deposited with the U.S. Copyright Office.
"""
"""
 An enumeration class for the message type field which describe what
 kind of data is being sent inside the Message
"""
from enum import Enum

__author__ = "Supriyo Chakraborty, Shalisha Witherspoon, Dean Steuer"


class MessageType(Enum):
    """
    Message types for communication between party and aggregator
    """
    MODEL_UPDATE = 1
    MODEL_HYPERPARAMETERS = 2
    MODEL_PARAMETERS = 3
    REQUEST_MODEL_HYPERPARAMETERS = 4
    REQUEST_MODEL_UPDATE = 5
    REGISTER = 6
    TRAIN = 7
    SAVE_MODEL = 8
    PREDICT_MODEL = 9
    EVAL_MODEL = 10
    ACK = 11
    SYNC_MODEL = 12
    STOP = 14
    ERROR_AUTH = 400



#  (C) Copyright IBM Corp. 2020.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#       http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ibm_watson_machine_learning.libs.repo.mlrepository import MetaNames
from ibm_watson_machine_learning.libs.repo.mlrepository import MetaProps
from ibm_watson_machine_learning.libs.repo.mlrepository import ModelArtifact
from ibm_watson_machine_learning.libs.repo.util.library_imports import LibraryChecker
from ibm_watson_machine_learning.libs.repo.base_constants import *
from .python_version import PythonVersion
import numpy as np
import json

lib_checker = LibraryChecker()

if lib_checker.installed_libs[SCIKIT]:
    from ibm_watson_machine_learning.libs.repo.mlrepositoryartifact.scikit_pipeline_reader import ScikitPipelineReader
    from ibm_watson_machine_learning.libs.repo.mlrepositoryartifact.xgboost_model_reader import XGBoostModelReader
    from sklearn.base import BaseEstimator
    from ibm_watson_machine_learning.libs.repo.mlrepositoryartifact.version_helper import ScikitVersionHelper, XGBoostVersionHelper

if lib_checker.installed_libs[XGBOOST]:
    from xgboost import XGBRegressor, XGBClassifier
    from ibm_watson_machine_learning.libs.repo.mlrepositoryartifact.version_helper import ScikitVersionHelper
    import sys
    import os
    try:
        sys.stderr = open(os.devnull, 'w')
        import xgboost as xgb
    except Exception as ex:
        print ('Failed to import xgboost. Error: ' + str(ex))
    finally:
        sys.stderr.close()  # close /dev/null
        sys.stderr = sys.__stderr__


class ScikitPipelineModelArtifact(ModelArtifact):
    """
    Class of model artifacts created with MLRepositoryCLient.

    :param sklearn.pipeline.Pipeline scikit_pipeline_model: Pipeline Model which will be wrapped
    """
    def __init__(self, scikit_pipeline_model, training_features=None, training_target=None, feature_names=None,
                 label_column_names=None, uid=None, name=None, meta_props=MetaProps({})):
        lib_checker.check_lib(SCIKIT)
        super(ScikitPipelineModelArtifact, self).__init__(uid, name, meta_props)

        is_scikit, is_xgboost = False, False

        if issubclass(type(scikit_pipeline_model), BaseEstimator):
            is_scikit = True

        if not is_scikit and lib_checker.installed_libs[XGBOOST]:
            if isinstance(scikit_pipeline_model, xgb.Booster):
                is_xgboost = True

        if not (is_scikit or is_xgboost):
            raise ValueError('Invalid type for scikit ml_pipeline_model: {}'.
                             format(scikit_pipeline_model.__class__.__name__))

        self.ml_pipeline_model = scikit_pipeline_model
        self.ml_pipeline = None     # no pipeline or parent reference


        if meta_props.prop(MetaNames.RUNTIMES) is None and meta_props.prop(MetaNames.RUNTIME_UID) is None and meta_props.prop(MetaNames.FRAMEWORK_RUNTIMES) is None:
            ver = PythonVersion.significant()
            runtimes = '[{"name":"python","version": "'+ ver + '"}]'
            self.meta.merge(
                MetaProps({MetaNames.FRAMEWORK_RUNTIMES: runtimes})
            )

        if is_xgboost:
            self.meta.merge(
                MetaProps({
                    MetaNames.FRAMEWORK_NAME: XGBoostVersionHelper.model_type(scikit_pipeline_model),
                    MetaNames.FRAMEWORK_VERSION: XGBoostVersionHelper.model_version(scikit_pipeline_model)
                })
            )

            if(training_features is not None):
                if(training_target is None):
                    if not (isinstance(training_features, xgb.DMatrix)):
                        raise ValueError("Training target column has not been provided for the training data set")
                self.meta.merge(self._get_schema(training_features, training_target, feature_names, label_column_names))

            self._reader = XGBoostModelReader(self.ml_pipeline_model)
        else:
            if lib_checker.installed_libs[XGBOOST]:
                if (issubclass(type(scikit_pipeline_model), XGBClassifier) or issubclass(type(scikit_pipeline_model), XGBRegressor)):
                    if MetaNames.FRAMEWORK_LIBRARIES not in self.meta.meta:
                        framewk_library_entry = {"name": "xgboost",
                                                 "version": xgb.__version__}
                        framewk_library_entry_full = json.dumps([framewk_library_entry])
                        self.meta.merge(
                            MetaProps({
                                MetaNames.FRAMEWORK_LIBRARIES: framewk_library_entry_full
                            })
                        )
                else:
                    self._xgboost_in_model(scikit_pipeline_model)

            self.meta.merge(
                MetaProps({
                    MetaNames.FRAMEWORK_NAME: ScikitVersionHelper.model_type(scikit_pipeline_model),
                    MetaNames.FRAMEWORK_VERSION: ScikitVersionHelper.model_version(scikit_pipeline_model)
                })
            )

            if(training_features is not None):
                if(training_target is None):
                    raise ValueError("Training target column has not been provided for the training data set")
                self.meta.merge(self._get_schema(training_features, training_target, feature_names, label_column_names))

            self._reader = ScikitPipelineReader(self.ml_pipeline_model)


    def _xgboost_in_model(self, model):
        import sklearn
        if '17' in sklearn.__version__:
            from sklearn.grid_search import GridSearchCV, RandomizedSearchCV
        else:
            from sklearn.model_selection import GridSearchCV, RandomizedSearchCV
        from sklearn.multiclass import OneVsRestClassifier
        from sklearn.pipeline import Pipeline
        from sklearn.ensemble import VotingClassifier
        _estimators = [GridSearchCV, RandomizedSearchCV, VotingClassifier, OneVsRestClassifier, Pipeline]

        set_libraries = False

        for estimator in _estimators:
            if isinstance(model, estimator):
                if isinstance(model, GridSearchCV) or isinstance(model, RandomizedSearchCV):
                    estimator = model.best_estimator_
                    if isinstance(estimator, XGBRegressor) or isinstance(estimator, XGBClassifier):
                        set_libraries = True
                elif isinstance(model, Pipeline):
                    estimator = model.steps[-1][1]
                    if isinstance(estimator, XGBRegressor) or isinstance(estimator, XGBClassifier):
                        set_libraries = True
                else:
                    for est in model.estimators_:
                        if isinstance(est, XGBRegressor) or isinstance(est, XGBClassifier):
                            set_libraries = True

        if set_libraries:
            if MetaNames.FRAMEWORK_LIBRARIES not in self.meta.meta:
                framewk_library_entry = {"name": "xgboost",
                                         "version": xgb.__version__}
                framewk_library_entry_full = json.dumps([framewk_library_entry])
                self.meta.merge(
                    MetaProps({
                        MetaNames.FRAMEWORK_LIBRARIES: framewk_library_entry_full
                    })
                )

    def _get_schema(self, training_features, training_target, feature_names, label_column_names):
        is_xgboost_dmatrix = False

        if(training_target is None):           #training_target is None for Dmatrix
            training_props = {
                # "features": {"type": type(training_features).__name__, "fields": []},
                # "labels": {"type": type(training_features.get_label()).__name__, "fields": []}
                "type": type(training_features).__name__,
                "fields" : [],
                "labels": { "fields": []}
            }
        else:
            training_props = {
                "type": type(training_features).__name__,
                "fields": [],
                "labels": {"fields": []}
            }

        lib_checker.check_lib(PANDAS)
        import pandas as pd

        # Check feature types, currently supporting pandas df, numpy.ndarray, python lists and Dmatrix
        if(isinstance(training_features, pd.DataFrame)):
            for feature in training_features.dtypes.iteritems():
                training_props["fields"].append({"name": feature[0], "type": str(feature[1])})
        elif(isinstance(training_features, np.ndarray)):
            dims = training_features.shape
            if len(dims) == 1:
                if feature_names is None:
                    feature_names = 'f1'
                training_props["fields"].append({"name": feature_names, "type": type(training_features[0]).__name__})
            else:
                if feature_names is None:
                    feature_names = ['f' + str(i) for i in range(dims[1])]
                elif isinstance(feature_names, np.ndarray):
                    feature_names = feature_names.tolist()
                for i in range(dims[1]):
                    training_props["fields"].append({
                        "name": feature_names[i], "type": type(training_features.item(0, i)).__name__
                    })

        elif(isinstance(training_features, list)):
            if not isinstance(training_features[0], list):
                if feature_names is None:
                    feature_names = 'f1'
                training_props["fields"].append({"name": feature_names, "type": type(training_features[0]).__name__})
            else:
                if feature_names is None:
                    feature_names = ['f' + str(i) for i in range(len(training_features[0]))]
                elif isinstance(feature_names, np.ndarray):
                    feature_names = feature_names.tolist()
                for i in range(len(training_features[0])):
                    training_props["fields"].append({
                        "name": feature_names[i], "type": type(training_features[0][i]).__name__
                    })
        elif(isinstance(training_features, xgb.DMatrix)):
            for index, feature in enumerate(training_features.feature_names):   #start=0
                training_props["fields"].append({"name": feature, "type": str(training_features.feature_types[index])})
            is_xgboost_dmatrix = True
        else:
            raise ValueError("Unsupported training data type %s provided" % (type(training_features).__name__))

        #Check target or label data types
        if(isinstance(training_target, pd.DataFrame)):
            for feature in training_target.dtypes.iteritems():
                training_props["labels"]["fields"].append({"name": feature[0], "type": str(feature[1])})
        elif(isinstance(training_target, pd.Series)):
            training_props["labels"]["fields"].append({"name": training_target.name, "type": str(training_target.dtype)})
        elif(isinstance(training_target, np.ndarray)):
            dims = training_target.shape
            if len(dims) == 1:
                if label_column_names is None:
                    label_column_names = 'l1'
                elif isinstance(label_column_names, list) or isinstance(label_column_names, np.ndarray):
                    label_column_names = label_column_names[0]
                training_props["labels"]["fields"].append({"name": label_column_names, "type": type(training_target.item(0)).__name__})
            else:
                if label_column_names is None:
                    label_column_names = ['l' + str(i) for i in range(dims[1])]
                elif isinstance(label_column_names, np.ndarray):
                    label_column_names = label_column_names.tolist()
                for i in range(dims[1]):
                    training_props["labels"]["fields"].append({
                        "name": label_column_names[i], "type": type(training_target.item(0, i)).__name__
                    })
        elif(isinstance(training_target, list)):
            if not isinstance(training_target[0], list):
                if label_column_names is None:
                    label_column_names = 'l1'
                elif isinstance(label_column_names, list) or isinstance(label_column_names, np.ndarray):
                    label_column_names = label_column_names[0]
                training_props["labels"]["fields"].append({
                    "name": label_column_names, "type": type(training_target[0]).__name__
                })
            else:
                if label_column_names is None:
                    label_column_names = ['l' + str(i) for i in range(len(training_target[0]))]
                elif isinstance(label_column_names, np.ndarray):
                    label_column_names = label_column_names.tolist()
                for i in range(len(training_target[0])):
                    training_props["labels"]["fields"].append({
                        "name": label_column_names[i], "type": type(training_target[0][i]).__name__
                    })
        elif(isinstance(training_features, xgb.DMatrix)):  #For DMatrix labels is part of training_data and not training_target
            training_props["labels"]["fields"].append({"type": str(type(training_features.get_label()[0]).__name__)})
        else:
            raise ValueError("Unsupported label data type %s provided" % (type(training_target)))

        if is_xgboost_dmatrix:
            label = None
        else:
            label = training_props["labels"]["fields"][0]["name"]

        del training_props["labels"]
        return MetaProps({
            MetaNames.TRAINING_DATA_SCHEMA: training_props,
            MetaNames.LABEL_FIELD: label
        })

    def pipeline_artifact(self):
        """
        Returns None. Pipeline is not implemented for scikit model.
        """
        pass

    def reader(self):
        """
        Returns reader used for getting pipeline model content.

        :return: reader for sklearn.pipeline.Pipeline
        :rtype: ScikitPipelineReader
        """

        return self._reader

    def _copy(self, uid=None, meta_props=None):
        if uid is None:
            uid = self.uid

        if meta_props is None:
            meta_props = self.meta

        return ScikitPipelineModelArtifact(
            self.ml_pipeline_model,
            uid=uid,
            name=self.name,
            meta_props=meta_props
        )


import unittest
from sklearn.pipeline import Pipeline
from pprint import pprint
import pandas as pd
import traceback
from os import environ

from ibm_watson_machine_learning import APIClient

from ibm_watson_machine_learning.experiment import AutoAI
from ibm_watson_machine_learning.deployment import Batch
from ibm_watson_machine_learning.preprocessing import DataJoinGraph
from ibm_watson_machine_learning.preprocessing.data_join_pipeline import OBMPipelineGraphBuilder

from ibm_watson_machine_learning.helpers.connections import S3Connection, S3Location, DataConnection, FSLocation
from ibm_watson_machine_learning.utils import WMLClientError

from ibm_watson_machine_learning.experiment.autoai.optimizers import RemoteAutoPipelines

from ibm_watson_machine_learning.tests.utils import bucket_exists, \
    create_bucket, get_space_id

from ibm_watson_machine_learning.tests.utils.cleanup import space_cleanup

from ibm_watson_machine_learning.tests.utils import (
    get_wml_credentials, get_cos_credentials, is_cp4d, print_test_separators)
from ibm_watson_machine_learning.utils.autoai.errors import WrongDataJoinGraphNodeName

get_step_details = OBMPipelineGraphBuilder.get_step_details
get_join_extractors = OBMPipelineGraphBuilder.get_join_extractors
get_extractor_columns = OBMPipelineGraphBuilder.get_extractor_columns
get_extractor_transformations = OBMPipelineGraphBuilder.get_extractor_transformations


class TestOBMOnly(unittest.TestCase):
    data_join_graph: 'DataJoinGraph' = None
    wml_client: 'APIClient' = None
    experiment: 'AutoAI' = None
    remote_auto_pipelines: 'RemoteAutoPipelines' = None
    wml_credentials = None
    cos_credentials = None
    pipeline_opt: 'RemoteAutoPipelines' = None
    historical_opt: 'RemoteAutoPipelines' = None
    service: 'Batch' = None
    data_join_pipeline: 'DataJoinPipeline' = None

    data_location = './autoai/data/'

    trained_pipeline_details = None
    run_id = None

    data_connections = []
    main_conn = None
    customers_conn = None
    transactions_conn = None
    purchases_conn = None
    products_conn = None
    results_connection = None

    train_data = None
    holdout_data = None
    bucket_name = environ.get('BUCKET_NAME', "wml-autoaitests-obm-qa")
    space_name = environ.get('SPACE_NAME', 'regression_tests_sdk_space')

    cos_endpoint =  "https://s3.us-south.cloud-object-storage.appdomain.cloud"
    cos_resource_instance_id = None
    data_cos_path = 'data/'

    results_cos_path = 'results_wml_autoai'

    pipeline: 'Pipeline' = None
    lale_pipeline = None
    deployed_pipeline = None
    hyperopt_pipelines = None
    new_pipeline = None
    new_sklearn_pipeline = None

    OPTIMIZER_NAME = 'OBM only test'
    DEPLOYMENT_NAME = "OBM only deployment test"

    job_id = None

    # # CP4D CONNECTION DETAILS:
    # if 'qa' in get_env().lower():
    #     space_id = 'aa81ee79-795a-4a23-8265-e87966dfe437'
    #     project_id = '3d177338-3f8b-444d-b385-9ff929fd124f'
    # else:
    #     project_id = '051e12c2-a65d-49d7-8ec3-04e3c36a60eb'
    #     space_id = None

    project_id = None
    space_id = None
    asset_id = None

    @classmethod
    def setUpClass(cls) -> None:
        """
        Load WML credentials from config.ini file based on ENV variable.
        """
        cls.wml_credentials = get_wml_credentials()

        if not is_cp4d():
            cls.cos_credentials = get_cos_credentials()
            if 'endpoint_url' in cls.cos_credentials:
                cls.cos_endpoint = cls.cos_credentials['endpoint_url']
            cls.cos_resource_instance_id = cls.cos_credentials.get('resource_instance_id')

        cls.wml_client = APIClient(wml_credentials=cls.wml_credentials.copy())
        cls.project_id = cls.wml_credentials.get('project_id')
        cls.wml_client.set.default_project(cls.project_id)

    # @print_test_separators
    # def test_00a_space_cleanup(self):
    #     space_cleanup(self.wml_client,
    #                   get_space_id(self.wml_client, self.space_name,
    #                                cos_resource_instance_id=self.cos_resource_instance_id),
    #                   days_old=7)
    #     TestOBMOnly.space_id = get_space_id(self.wml_client, self.space_name,
    #                                  cos_resource_instance_id=self.cos_resource_instance_id)



    @print_test_separators
    def test_00b_prepare_COS_instance(self):
        if self.wml_client.ICP or self.wml_client.WSD:
            self.skipTest("Prepare COS is available only for Cloud")

        import ibm_boto3
        cos_resource = ibm_boto3.resource(
            service_name="s3",
            endpoint_url=self.cos_endpoint,
            aws_access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
            aws_secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key']
        )
        # Prepare bucket
        if not bucket_exists(cos_resource, self.bucket_name):
            TestOBMOnly.bucket_name = create_bucket(cos_resource, self.bucket_name)

            self.assertIsNotNone(TestOBMOnly.bucket_name)
            self.assertTrue(bucket_exists(cos_resource, TestOBMOnly.bucket_name))

        print(f"Using COS bucket: {TestOBMOnly.bucket_name}")

    @print_test_separators
    def test_01_create_multiple_data_connections__connections_created(self):
        print("Creating multiple data connections...")
        if self.wml_client.ICP:
            TestOBMOnly.main_conn = DataConnection(
                data_join_node_name="main",
                location=FSLocation(path=self.data_location + 'group_customer_main.csv'))

            TestOBMOnly.customers_conn = DataConnection(
                data_join_node_name="customers",
                location=FSLocation(path=self.data_location + 'group_customer_customers.csv'))

            TestOBMOnly.transactions_conn = DataConnection(
                data_join_node_name="transactions",
                location=FSLocation(path=self.data_location + 'group_customer_transactions.csv'))

            TestOBMOnly.purchases_conn = DataConnection(
                data_join_node_name="purchases",
                location=FSLocation(path=self.data_location + 'group_customer_purchase.csv'))

            TestOBMOnly.products_conn = DataConnection(
                data_join_node_name="products",
                location=FSLocation(path=self.data_location + 'group_customer_products.csv'))

            TestOBMOnly.results_connection = None
        else:
            TestOBMOnly.main_conn = DataConnection(
                data_join_node_name="main",
                connection=S3Connection(
                    endpoint_url=self.cos_endpoint,
                    access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
                    secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key']),
                location=S3Location(bucket=self.bucket_name,
                                    path=self.data_cos_path + 'group_customer_main.csv'))

            TestOBMOnly.customers_conn = DataConnection(
                data_join_node_name="customers",
                connection=S3Connection(
                    endpoint_url=self.cos_endpoint,
                    access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
                    secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key']),
                location=S3Location(bucket=self.bucket_name,
                                    path=self.data_cos_path + 'group_customer_customers.csv'))

            TestOBMOnly.transactions_conn = DataConnection(
                data_join_node_name="transactions",
                connection=S3Connection(
                    endpoint_url=self.cos_endpoint,
                    access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
                    secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key']),
                location=S3Location(bucket=self.bucket_name,
                                    path=self.data_cos_path + 'group_customer_transactions.csv'))

            TestOBMOnly.purchases_conn = DataConnection(
                data_join_node_name="purchases",
                connection=S3Connection(
                    endpoint_url=self.cos_endpoint,
                    access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
                    secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key']),
                location=S3Location(bucket=self.bucket_name,
                                    path=self.data_cos_path + 'group_customer_purchase.csv'))

            TestOBMOnly.products_conn = DataConnection(
                data_join_node_name="products",
                connection=S3Connection(
                    endpoint_url=self.cos_endpoint,
                    access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
                    secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key']),
                location=S3Location(bucket=self.bucket_name,
                                    path=self.data_cos_path + 'group_customer_products.csv'))

            TestOBMOnly.results_connection = DataConnection(
                connection=S3Connection(endpoint_url=self.cos_endpoint,
                                        access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
                                        secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key']),
                location=S3Location(bucket=self.bucket_name,
                                    path=self.results_cos_path)
            )

        TestOBMOnly.data_connections = [self.main_conn,
                                        self.customers_conn,
                                        self.transactions_conn,
                                        self.purchases_conn,
                                        self.products_conn]

    @print_test_separators
    def test_02_create_data_join_graph__graph_created(self):
        print("Defining DataJoinGraph...")
        data_join_graph = DataJoinGraph()
        data_join_graph.node(name="main")
        data_join_graph.node(name="customers")
        data_join_graph.node(name="transactions")
        data_join_graph.node(name="purchases")
        data_join_graph.node(name="products")
        data_join_graph.edge(from_node="main", to_node="customers",
                             from_column=["group_customer_id"], to_column=["group_customer_id"])
        data_join_graph.edge(from_node="main", to_node="transactions",
                             from_column=["transaction_id"], to_column=["transaction_id"])
        data_join_graph.edge(from_node="main", to_node="purchases",
                             from_column=["group_id"], to_column=["group_id"])
        data_join_graph.edge(from_node="transactions", to_node="products",
                             from_column=["product_id"], to_column=["product_id"])

        TestOBMOnly.data_join_graph = data_join_graph

        print(f"data_join_graph: {data_join_graph}")

    # @print_test_separators
    # def test_03_data_join_graph_visualization(self):
    #     print("Visualizing data_join_graph...")
    #     self.data_join_graph.visualize()

    @print_test_separators
    def test_04_initialize_AutoAI_experiment__pass_credentials__object_initialized(self):
        print("Initializing AutoAI experiment...")
        TestOBMOnly.experiment = AutoAI(wml_credentials=self.wml_credentials.copy(),
                                        project_id=self.project_id)

        self.assertIsInstance(self.experiment, AutoAI, msg="Experiment is not of type AutoAI.")

    @print_test_separators
    def test_05_save_remote_data_and_DataConnection_setup(self):
        print("Writing multiple data files to COS...")
        if is_cp4d():
            pass

        else:  # for cloud and COS
            self.main_conn.write(data=self.data_location + 'group_customer_main.csv',
                                 remote_name=self.data_cos_path + 'group_customer_main.csv')
            self.customers_conn.write(data=self.data_location + 'group_customer_customers.csv',
                                      remote_name=self.data_cos_path + 'group_customer_customers.csv')
            self.transactions_conn.write(data=self.data_location + 'group_customer_transactions.csv',
                                         remote_name=self.data_cos_path + 'group_customer_transactions.csv')
            self.purchases_conn.write(data=self.data_location + 'group_customer_purchase.csv',
                                      remote_name=self.data_cos_path + 'group_customer_purchase.csv')
            self.products_conn.write(data=self.data_location + 'group_customer_products.csv',
                                     remote_name=self.data_cos_path + 'group_customer_products.csv')

    @print_test_separators
    def test_06_initialize_optimizer(self):
        print("Initializing optimizer with data_join_graph...")
        TestOBMOnly.remote_auto_pipelines = self.experiment.optimizer(
            name=self.OPTIMIZER_NAME,
            prediction_type=AutoAI.PredictionType.REGRESSION,
            prediction_column='next_purchase',
            scoring=AutoAI.Metrics.LOG_LOSS,
            t_shirt_size=self.experiment.TShirtSize.L if not is_cp4d() else self.experiment.TShirtSize.S,
            data_join_graph=self.data_join_graph,
            data_join_only=True,
        )

        self.assertIsInstance(self.remote_auto_pipelines, RemoteAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

    @print_test_separators
    def test_07_get_configuration_parameters_of_remote_auto_pipeline(self):
        print("Getting experiment configuration parameters...")
        parameters = self.remote_auto_pipelines.get_params()
        print(parameters)
        self.assertIsInstance(parameters, dict, msg='Config parameters are not a dictionary instance.')

    def test_08a_check_data_join_node_names(self):
        if self.wml_client.ICP:
            main_conn = DataConnection(
                location=FSLocation(path=self.data_location + 'group_customer_main.csv'))

            customers_conn = DataConnection(
                location=FSLocation(path=self.data_location + 'group_customer_customers.csv'))

            transactions_conn = DataConnection(
                location=FSLocation(path=self.data_location + 'group_customer_transactions.csv'))

            purchases_conn = DataConnection(
                location=FSLocation(path=self.data_location + 'group_customer_purchase.csv'))

            products_conn = DataConnection(
                location=FSLocation(path=self.data_location + 'group_customer_products.csv'))
        else:
            main_conn = DataConnection(
                connection=S3Connection(
                    endpoint_url=self.cos_endpoint,
                    access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
                    secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key']),
                location=S3Location(bucket=self.bucket_name,
                                    path=self.data_cos_path + 'group_customer_main.csv'))

            customers_conn = DataConnection(
                connection=S3Connection(
                    endpoint_url=self.cos_endpoint,
                    access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
                    secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key']),
                location=S3Location(bucket=self.bucket_name,
                                    path=self.data_cos_path + 'group_customer_customers.csv'))

            transactions_conn = DataConnection(
                connection=S3Connection(
                    endpoint_url=self.cos_endpoint,
                    access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
                    secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key']),
                location=S3Location(bucket=self.bucket_name,
                                    path=self.data_cos_path + 'group_customer_transactions.csv'))

            purchases_conn = DataConnection(
                connection=S3Connection(
                    endpoint_url=self.cos_endpoint,
                    access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
                    secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key']),
                location=S3Location(bucket=self.bucket_name,
                                    path=self.data_cos_path + 'group_customer_purchase.csv'))

            products_conn = DataConnection(
                connection=S3Connection(
                    endpoint_url=self.cos_endpoint,
                    access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
                    secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key']),
                location=S3Location(bucket=self.bucket_name,
                                    path=self.data_cos_path + 'group_customer_products.csv'))

        data_connections = [main_conn,
                            customers_conn,
                            transactions_conn,
                            purchases_conn,
                            products_conn]

        with self.assertRaises(WrongDataJoinGraphNodeName, msg="Training started with wrong data join node names"):
            self.remote_auto_pipelines.fit(
                training_data_reference=data_connections,
                training_results_reference=self.results_connection,
                background_mode=False)

    @print_test_separators
    def test_08b_fit_run_training_of_auto_ai_in_wml(self):
        print("Scheduling OBM only training...")
        TestOBMOnly.trained_pipeline_details = self.remote_auto_pipelines.fit(
            training_data_reference=self.data_connections,
            training_results_reference=self.results_connection,
            background_mode=False)

        TestOBMOnly.run_id = self.trained_pipeline_details['metadata']['id']

        for connection in self.data_connections:
            self.assertIsNotNone(
                connection.auto_pipeline_params,
                msg=f'DataConnection auto_pipeline_params was not updated for connection: {connection.id}')

    @print_test_separators
    def test_09_download_original_training_data(self):
        print("Downloading each training file...")
        for connection in self.remote_auto_pipelines.get_data_connections():
            train_data = connection.read()

            print(f"Connection: {connection.id} - train data sample:")
            print(train_data.head())
            self.assertGreater(len(train_data), 0)

    @print_test_separators
    def test_10_download_preprocessed_obm_training_data(self):
        print("Downloading OBM preprocessed training data with holdout split...")
        TestOBMOnly.train_data, TestOBMOnly.holdout_data = (
            self.remote_auto_pipelines.get_preprocessed_data_connection().read(with_holdout_split=True))

        print("OBM train data sample:")
        print(self.train_data.head())
        self.assertGreater(len(self.train_data), 0)
        print("OBM holdout data sample:")
        print(self.holdout_data.head())
        self.assertGreater(len(self.holdout_data), 0)

    @print_test_separators
    def test_11_visualize_obm_pipeline(self):
        if self.wml_client.ICP or self.wml_client.WSD:
            self.skipTest("Visualizing pipeline not supported on CPD")

        print("Visualizing OBM model ...")
        TestOBMOnly.data_join_pipeline = self.remote_auto_pipelines.get_preprocessing_pipeline()
        assert isinstance(self.data_join_pipeline._pipeline_json, list)
        assert '40 [label=NumberSet]' in self.data_join_pipeline._graph.__str__()
    #     self.data_join_pipeline.visualize()

    @print_test_separators
    def test_12_check_if_data_join_pipeline_graph_correct(self):
        if self.wml_client.ICP or self.wml_client.WSD:
            self.skipTest("Visualizing pipeline not supported on CPD")
        pipeline_json = self.data_join_pipeline._pipeline_json
        graph = self.data_join_pipeline._graph_json

        step_types = [message['feature_engineering_components']['obm'][0]['step_type'] for message in pipeline_json]
        last_non_join_iteration = step_types.index('join')
        selection_iteration = step_types.index('feature selection') + 1
        join_iterations = [i + 1 for i, x in enumerate(step_types) if x == "join"]

        for message in pipeline_json:
            name, iteration, _ = get_step_details(message)
            self.assertTrue(str(iteration) in graph['nodes'])

            if 1 < iteration <= 2:
                self.assertTrue(str(iteration) in graph['edges'][str(iteration - 1)])
            elif iteration in join_iterations:
                self.assertTrue(str(iteration) in graph['edges'][str(last_non_join_iteration)])
                extractors = get_join_extractors(message)

                if extractors is None:
                    continue
                for ext, i in zip(extractors, range(len(extractors))):
                    ext_index = str(iteration) + str(i)
                    self.assertTrue(ext_index in graph['nodes'] and
                                    ext_index in ext_index in graph['edges'][str(iteration)])

                    columns = get_extractor_columns(extractors[ext])
                    transformations = get_extractor_transformations(extractors[ext])
                    for j, column in enumerate(columns):
                        col_index = str(iteration) + str(i) + str(j)
                        self.assertTrue(col_index in graph['nodes'] and col_index in graph['edges'][str(ext_index)])

                        for transformation in transformations:
                            self.assertTrue(transformation in graph['edges'][str(col_index)])
                            self.assertTrue(str(selection_iteration) in graph['edges'][str(transformation)])

            elif iteration > selection_iteration:
                self.assertTrue(str(iteration) in graph['edges'][str(iteration - 1)])

    @print_test_separators
    def test_13_get_run_status(self):
        print("Getting training status...")
        status = self.remote_auto_pipelines.get_run_status()
        print(status)
        self.assertEqual("completed", status, msg="AutoAI run didn't finished successfully. Status: {}".format(status))

    @print_test_separators
    def test_14_get_run_details(self):
        print("Getting training details...")
        parameters = self.remote_auto_pipelines.get_run_details()
        print(parameters)
        self.assertIsNotNone(parameters)

    def test_14b_get_metrics(self):
        with self.assertRaises(WMLClientError, msg="No metrics available for obm only"):
            self.wml_client.training.get_metrics(self.run_id)

    @print_test_separators
    def test_15__get_data_connections__return_a_list_with_data_connections_with_optimizer_params(self):
        print("Getting all data connections...")
        data_connections = self.remote_auto_pipelines.get_data_connections()
        self.assertIsInstance(data_connections, list, msg="There should be a list container returned")
        self.assertIsInstance(data_connections[0], DataConnection,
                              msg="There should be a DataConnection object returned")

    @print_test_separators
    def test_16_get_historical_optimizer_with_data_join_graph(self):
        print("Fetching historical optimizer with OBM only...")
        TestOBMOnly.historical_opt = self.experiment.runs.get_optimizer(run_id=self.run_id)
        self.assertIsInstance(self.historical_opt.get_params()['data_join_graph'], DataJoinGraph,
                              msg="data_join_graph was incorrectly loaded")

    @print_test_separators
    def test_17_get_obm_training_data_from_historical_optimizer(self):
        print("Download historical preprocessed data from OBM stage...")
        train_data = self.historical_opt.get_preprocessed_data_connection().read()

        print("OBM train data sample:")
        print(train_data.head())
        self.assertGreater(len(train_data), 0)


if __name__ == '__main__':
    unittest.main()
